"""
ttfpdf v1.4.3 May 30 2020
provides support for the proofpdf script,  for working with OpenType/TTF
fonts. Provides an implementation of the fontpdf font object. Cannot be
run alone.
"""
__copyright__ = """Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
"""

from fontTools.pens.boundsPen import BoundsPen
from afdko.pdflib.fontpdf import FontPDFGlyph, FontPDFFont, FontPDFPen

__version__ = "1.4.3"


class txPDFFont(FontPDFFont):

	def __init__(self, clientFont, params):
		self.clientFont = clientFont
		if params.userBaseLine != None:
			self.baseLine = params.userBaseLine
		else:
			self.baseLine = None
		self.path = params.rt_filePath
		self.isCID = 0
		self.psName = None;
		self.OTVersion = None;
		self.emSquare = None
		self.bbox = None
		self.ascent = None
		self.descent = None
		self.blueZones = None
		self.getEmSquare()
		self.getBaseLine()
		self.getBBox()
		self.GetBlueZones()
		self.AscentDescent()
		return

	def clientGetPSName(self):
		psName = 'None'
		name_tb = self.clientFont['name']
		mac_name_id_6 = name_tb.getName(6, 1, 0, 0)
		win_name_id_6 = name_tb.getName(6, 3, 1, 1033)
		if mac_name_id_6:
			return mac_name_id_6.string.decode()
		elif win_name_id_6:
			return win_name_id_6.string.decode('utf_16_be')
		return psName

	def clientGetOTVersion(self):
		version =  self.clientFont['head'].fontRevision
		majorVersion = int(version)
		minorVersion = str(int( 1000*(0.0005 + version -majorVersion) )).zfill(3)
		versionString = "%s.%s" % (majorVersion, minorVersion)
		#print versionString
		return versionString

	def clientGetGlyph(self, glyphName):
		return txPDFGlyph(self, glyphName)

	def clientGetEmSquare(self):
		emSquare =  self.clientFont['head'].unitsPerEm
		return emSquare

	def clientGetBaseline(self):
		baseLine = 0
		txFont = self.clientFont
		try:
			unicodeRange2 = txFont["OS/2"].ulUnicodeRange2
			if unicodeRange2 & 0x10000: # supports CJK  ideographs
				baseTag = "ideo"
			else:
				baseTag = "romn"
			baseTable =  self.clientFont['BASE']
			baseTagIndex = baseTable.table.HorizAxis.BaseTagList.BaselineTag.index( baseTag)
			baseScript = None
			for baseRecord in baseTable.table.HorizAxis.BaseScriptList.BaseScriptRecord:
				if baseRecord.BaseScriptTag == "latn":
					baseScript = baseRecord.BaseScript
					baseLine = baseScript.BaseValues.BaseCoord[baseTagIndex].Coordinate
					break
		except (KeyError, AttributeError):
			baseLine = 0

		return baseLine

	def clientGetBBox(self):
		headTable =  self.clientFont['head']
		return [headTable.xMin, headTable.yMin, headTable.xMax, headTable.yMax]

	def clientGetBlueZones(self):
		blueValues = [[]]
		return blueValues
	def clientGetAscentDescent(self):
		txFont = self.clientFont
		try:
			os2Table =  self.clientFont['OS/2']
			return os2Table.sTypoAscender, os2Table.sTypoDescender
		except KeyError:
			return None, None

class  txPDFGlyph(FontPDFGlyph):

	def clientInitData(self):
		self.isTT = 1
		self.isCID = 0
		txFont = self.parentFont.clientFont
		glyphSet = txFont.getGlyphSet(preferCFF=False)
		clientGlyph = glyphSet[self.name]
		# Get the list of points
		pen = FontPDFPen(glyphSet)
		clientGlyph.draw(pen)

		if not hasattr(txFont, 'vmetrics'):
			try:
				txFont.vmetrics = txFont['vmtx'].metrics
			except KeyError:
				txFont.vmetrics = None
			try:
				txFont.vorg = txFont['VORG']
			except KeyError:
				txFont.vorg = None

		self.hhints = []
		self.vhints =[]
		self.numMT = pen.numMT
		self.numLT = pen.numLT
		self.numCT = pen.numCT
		self.numPaths = pen.numPaths
		self.pathList = pen.pathList
		for path in self.pathList :
			lenPath = len(path)
			path[-1].next = path[0]
			path[0].last = path[-1]
			if lenPath > 1:
				path[0].next = path[1]
				path[-1].last = path[-2]
				for i in range(lenPath)[1:-1]:
					pt = path[i]
					pt.next =  path[i+1]
					pt.last =  path[i-1]

		assert len(self.pathList) == self.numPaths, " Path lengths don't match %s %s" % (len(self.pathList) , self.numPaths)
		# get the bbox and width.
		pen = BoundsPen(glyphSet)
		clientGlyph.draw(pen)
		self.xAdvance = clientGlyph.width
		glyph_bounds = pen.bounds
		if not glyph_bounds:
			self.BBox = [0, 0, 0, 0]
		else:
			self.BBox = [round(item) for item in glyph_bounds]

		self.yOrigin = self.parentFont.emSquare + self.parentFont.getBaseLine()
		if txFont.vorg:
			try:
				self.yOrigin  = txFont.vorg[self.name]
			except KeyError:
				if txFont.vmetrics:
					try:
						mtx = txFont.vmetrics[self.name]
						self.yOrigin = mtx[1] + self.BBox[3]
					except KeyError:
						pass

		haveVMTX = 0
		if txFont.vmetrics:
			try:
				mtx = txFont.vmetrics[self.name]
				self.yAdvance = mtx[0]
				self.tsb = mtx[1]
				haveVMTX =1
			except KeyError:
				pass
		if not haveVMTX:
			self.yAdvance = self.parentFont.getEmSquare()
			self.tsb = self.yOrigin - self.BBox[3] + self.parentFont.getBaseLine()

		# Get the fdIndex, so we can laterdetermine which set of blue values to use.
		self.fdIndex = 0
		return






import os
import pytest
from shutil import copy2, copytree, rmtree
import subprocess
import sys

from afdko.makeotf import (
    checkIfVertInFeature,
    getOptions,
    MakeOTFParams,
    getSourceGOADBData,
    readOptionFile,
    writeOptionsFile,
    kMOTFOptions,
    kOptionNotSeen,
)
from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
)
from test_utils import (
    get_input_path,
    get_expected_path,
    get_font_revision,
    generate_ttx_dump,
    font_has_table,
)
from runner import main as runner
from differ import main as differ, SPLIT_MARKER

# The following ensures that the current working directory is consistently set.
# This is necessary to avoid a problem that was discovered when running tests
# with cibuildwheel, which initiates the pytest command from root, which causes
# some minor discrepancies in relative vs absolute paths.
test_dir = os.path.dirname(os.path.realpath(__file__))
afdko_dir = os.path.join(test_dir, "..")
os.chdir(afdko_dir)

TOOL = 'makeotf'
CMD = ['-t', TOOL]

T1PFA_NAME = 't1pfa.pfa'
UFO2_NAME = 'ufo2.ufo'
UFO3_NAME = 'ufo3.ufo'
CID_NAME = 'cidfont.ps'
TTF_NAME = 'font.ttf'
OTF_NAME = 'SourceSans-Test.otf'

DATA_DIR = os.path.join(os.path.split(__file__)[0], TOOL + '_data')
TEMP_DIR = os.path.join(DATA_DIR, 'temp_output')


xfail_win = pytest.mark.xfail(
    sys.platform == 'win32',
    reason="Console's encoding is not UTF-8 ?")


xfail_py3_win = pytest.mark.xfail(
    sys.version_info >= (3, 0) and sys.platform == 'win32',
    reason="?")


def setup_module():
    """
    Create the temporary output directory
    """
    rmtree(TEMP_DIR, ignore_errors=True)
    os.mkdir(TEMP_DIR)


def teardown_module():
    """
    teardown the temporary output directory
    """
    rmtree(TEMP_DIR)


# -----
# Tests
# -----

def test_exit_no_option():
    # It's valid to run 'makeotf' without using any options,
    # but if a default-named font file is NOT found in the
    # current directory, the tool exits with an error
    with pytest.raises(subprocess.CalledProcessError) as err:
        subprocess.check_call([TOOL])
    assert err.value.returncode == 2


@pytest.mark.parametrize('arg', ['-v', '-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-j', '--bogus'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 2


@pytest.mark.parametrize('arg, input_filename, ttx_filename', [
    ([], T1PFA_NAME, 't1pfa-dev.ttx'),
    ([], UFO2_NAME, 'ufo2-dev.ttx'),
    ([], UFO3_NAME, 'ufo3-dev.ttx'),
    ([], CID_NAME, 'cidfont-dev.ttx'),
    ([], TTF_NAME, 'ttf-dev.ttx'),
    (['r'], T1PFA_NAME, 't1pfa-rel.ttx'),
    (['r'], UFO2_NAME, 'ufo2-rel.ttx'),
    (['r'], UFO3_NAME, 'ufo3-rel.ttx'),
    (['r'], CID_NAME, 'cidfont-rel.ttx'),
    (['r', 'gf'], TTF_NAME, 'ttf-rel.ttx'),
])
def test_input_formats(arg, input_filename, ttx_filename):
    if 'gf' in arg:
        arg.append(f'_{get_input_path("GOADB.txt")}')
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'o', f'_{actual_path}'] + arg)
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])


@pytest.mark.parametrize('args, ttx_fname', [
    ([], 'font_dev'),
    (['r'], 'font_rel'),
])
def test_build_font_and_check_messages(args, ttx_fname):
    actual_path = get_temp_file_path()
    expected_msg_path = get_expected_path(f'{ttx_fname}_output.txt')
    ttx_filename = f'{ttx_fname}.ttx'
    stderr_path = runner(CMD + [
        '-s', '-e', '-o', 'f', f'_{get_input_path("font.pfa")}',
                          'o', f'_{actual_path}'] + args)
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])
    assert differ([expected_msg_path, stderr_path,
                   '-r', r'^Built (development|release) mode font'])


def test_getSourceGOADBData():
    ttf_path = get_input_path('font.ttf')
    assert getSourceGOADBData(ttf_path) == [['.notdef', '.notdef', ''],
                                            ['a', 'a', 'uni0041'],
                                            ['g2', 'g2', '']]


@xfail_win
@pytest.mark.parametrize('input_filename', [
    T1PFA_NAME, UFO2_NAME, UFO3_NAME, CID_NAME])
def test_path_with_non_ascii_chars_bug222(input_filename):
    temp_dir = get_temp_dir_path('á意ê  ï薨õ 巽ù')
    os.makedirs(temp_dir)
    assert os.path.isdir(temp_dir)
    input_path = get_input_path(input_filename)
    temp_path = os.path.join(temp_dir, input_filename)
    if os.path.isdir(input_path):
        copytree(input_path, temp_path)
    else:
        copy2(input_path, temp_path)
    expected_path = os.path.join(temp_dir, OTF_NAME)
    assert os.path.exists(expected_path) is False
    runner(CMD + ['-o', 'f', f'_{temp_path}'])
    assert os.path.isfile(expected_path)


def test_font_with_hash_bug239():
    input_path = get_input_path('bug239/font.ufo')
    output_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{input_path}',
                        'o', f'_{output_path}'])
    assert font_has_table(output_path, 'CFF ')


def test_font_with_outdated_hash_bug239():
    input_path = get_input_path('bug239/font_outdated_hash.ufo')
    output_path = get_temp_file_path()
    with pytest.raises(subprocess.CalledProcessError) as err:
        runner(CMD + ['-o', 'f', f'_{input_path}',
                            'o', f'_{output_path}'])
    assert err.value.returncode == 2


@xfail_py3_win
@pytest.mark.parametrize('input_filename', [UFO2_NAME, UFO3_NAME])
def test_ufo_with_trailing_slash_bug280(input_filename):
    # makeotf will now save the OTF alongside the UFO instead of inside of it
    ufo_path = get_input_path(input_filename)
    temp_dir = get_temp_dir_path()
    tmp_ufo_path = os.path.join(temp_dir, input_filename)
    copytree(ufo_path, tmp_ufo_path)
    runner(CMD + ['-o', 'f', f'_{tmp_ufo_path}{os.sep}'])
    expected_path = os.path.join(temp_dir, OTF_NAME)
    assert os.path.isfile(expected_path)


@pytest.mark.parametrize('input_filename', [
    T1PFA_NAME, UFO2_NAME, UFO3_NAME, CID_NAME])
def test_output_is_folder_only_bug281(input_filename):
    # makeotf will output a default-named font to the folder
    input_path = get_input_path(input_filename)
    temp_dir = get_temp_dir_path()
    expected_path = os.path.join(temp_dir, OTF_NAME)
    assert os.path.exists(expected_path) is False
    runner(CMD + ['-o', 'f', f'_{input_path}',
                        'o', f'_{temp_dir}'])
    assert os.path.isfile(expected_path)


@pytest.mark.parametrize('input_filename', [
    't1pfa-noPSname.pfa',
    'ufo2-noPSname.ufo',
    'ufo3-noPSname.ufo',
    'cidfont-noPSname.ps',
])
def test_no_postscript_name_bug282(input_filename):
    # makeotf will fail for both UFO and Type 1 inputs
    with pytest.raises(subprocess.CalledProcessError) as err:
        runner(CMD + ['-o', 'f', f'_{input_filename}'])
    assert err.value.returncode == 2


@pytest.mark.parametrize('fea_filename, result', [
    (None, 0),  # No feature path at...
    ('missing_file', 0),  # No feature path at...
    ('feat0.fea', 0),
    ('feat1.fea', 1),
    ('feat2.fea', 0),  # Could not find the include file...
    ('feat3.fea', 0),  # Could not find the include file...
    ('feat4.fea', 1),
])
def test_find_vert_feature_bug148(fea_filename, result):
    fea_path = None
    if fea_filename:
        input_dir = get_input_path('bug148')
        fea_path = os.path.join(input_dir, fea_filename)
    assert checkIfVertInFeature(fea_path) == result


@pytest.mark.parametrize('args, result', [
    # 'result' corresponds to the values of the
    # options 'ReleaseMode' and 'SuppressHintWarnings'
    ([], (None, None)),
    (['-r'], ('true', None)),
    (['-shw'], (None, None)),
    (['-nshw'], (None, 'true')),
    (['-r', '-shw'], ('true', None)),
    (['-r', '-nshw'], ('true', 'true')),
    # makeotf option parsing has no mechanism for mutual exclusivity,
    # so the last option typed on the command line wins
    (['-shw', '-nshw'], (None, 'true')),
    (['-nshw', '-shw'], (None, None)),
])
def test_options_shw_nshw_bug457(args, result):
    params = MakeOTFParams()
    getOptions(params, args)
    assert params.opt_ReleaseMode == result[0]
    assert params.opt_SuppressHintWarnings == result[1]


@pytest.mark.parametrize('args, result', [
    # 'result' corresponds to the values of the
    # options 'MacCmapScriptID' and 'MacCmapLanguageID'
    # CJK MacCmapScriptIDs: Japan/1, CN/2, Korea/3, GB/25
    ([], (None, None)),
    (['-cs', '1'], ('1', None)),
    (['-cl', '2'], (None, '2')),
    (['-cs', '4', '-cl', '5'], ('4', '5')),
])
def test_options_cs_cl_bug459(args, result):
    params = MakeOTFParams()
    getOptions(params, args)
    assert params.opt_MacCmapScriptID == result[0]
    assert params.opt_MacCmapLanguageID == result[1]


def test_writeOptionsFile():
    actual_path = get_temp_file_path()
    expected_path = get_expected_path('proj_write.txt')
    params = MakeOTFParams()
    params.currentDir = os.path.dirname(actual_path)
    params.opt_InputFontPath = 'font.pfa'
    params.opt_OutputFontPath = 'fôñt.otf'
    params.opt_kSetfsSelectionBitsOn = [7]
    params.opt_kSetfsSelectionBitsOff = [9, 8]
    params.opt_ConvertToCID = 'true'
    params.opt_kSetOS2Version = 3
    params.verbose = True
    # set options's sort order
    kMOTFOptions['InputFontPath'][0] = 1
    kMOTFOptions['OutputFontPath'][0] = kOptionNotSeen + 1
    kMOTFOptions['kSetfsSelectionBitsOn'][0] = 30
    writeOptionsFile(params, actual_path)
    assert differ([expected_path, actual_path])


def test_readOptionFile():
    INPUT_FONT_NAME = 'font.pfa'
    OUTPUT_FONT_NAME = 'font.otf'
    proj_path = get_input_path('proj.txt')
    abs_font_dir_path = os.path.dirname(proj_path)
    params = MakeOTFParams()
    assert params.fontDirPath == '.'

    params.currentDir = os.path.dirname(proj_path)
    assert readOptionFile(proj_path, params, 1) == (True, 7)
    assert params.fontDirPath == '.'
    assert params.opt_InputFontPath == INPUT_FONT_NAME
    assert params.opt_OutputFontPath == OUTPUT_FONT_NAME
    assert params.opt_ConvertToCID == 'true'
    assert params.opt_kSetfsSelectionBitsOff == '[8, 9]'
    assert params.opt_kSetfsSelectionBitsOn == '[7]'
    assert params.seenOS2v4Bits == [1, 1, 1]
    assert params.opt_UseOldNameID4 is None

    params.currentDir = os.getcwd()
    font_dir_path = os.path.relpath(abs_font_dir_path, params.currentDir)
    input_font_path = os.path.join(font_dir_path, INPUT_FONT_NAME)
    output_font_path = os.path.join(font_dir_path, OUTPUT_FONT_NAME)
    assert readOptionFile(proj_path, params, 1) == (True, 7)
    assert params.fontDirPath == os.path.normpath(font_dir_path)
    assert params.opt_InputFontPath == os.path.normpath(input_font_path)
    assert params.opt_OutputFontPath == os.path.normpath(output_font_path)


def test_readOptionFile_abspath():
    proj_path = get_input_path('proj2.txt')
    params = MakeOTFParams()

    root = os.path.abspath(os.sep)
    params.currentDir = os.path.join(root, 'different_dir')
    assert readOptionFile(proj_path, params, 1) == (False, 3)
    assert params.fontDirPath.startswith(root)
    assert params.opt_InputFontPath.startswith(root)
    assert params.opt_OutputFontPath.startswith(root)


@pytest.mark.parametrize('cur_dir_str', [
    'different_dir',
    './different_dir',
    '../different_dir',
])
def test_readOptionFile_relpath(cur_dir_str):
    INPUT_FONT_NAME = 'font.pfa'
    OUTPUT_FONT_NAME = 'font.otf'
    proj_path = get_input_path('proj2.txt')
    abs_font_dir_path = os.path.dirname(proj_path)
    params = MakeOTFParams()

    if '/' in cur_dir_str:
        # flip the slashes used in the test's input string
        params.currentDir = os.path.normpath(cur_dir_str)
    else:
        params.currentDir = cur_dir_str

    font_dir_path = os.path.relpath(abs_font_dir_path, params.currentDir)

    if cur_dir_str.startswith('..') and os.path.dirname(os.getcwd()) == os.sep:
        # the project is inside a folder located at the root level;
        # remove the two dots at the start of the path, otherwise
        # testing input '../different_dir' will fail.
        font_dir_path = font_dir_path[2:]

    input_font_path = os.path.join(font_dir_path, INPUT_FONT_NAME)
    output_font_path = os.path.join(font_dir_path, OUTPUT_FONT_NAME)
    assert readOptionFile(proj_path, params, 1) == (False, 3)
    assert params.fontDirPath == os.path.normpath(font_dir_path)
    assert params.opt_InputFontPath == os.path.normpath(input_font_path)
    assert params.opt_OutputFontPath == os.path.normpath(output_font_path)


def test_readOptionFile_filenotfound():
    proj_path = get_input_path('notafile')
    params = MakeOTFParams()
    params.currentDir = os.getcwd()
    assert readOptionFile(proj_path, params, 0) == (True, 0)


@pytest.mark.parametrize('args, input_filename, ttx_filename', [
    (['r'], T1PFA_NAME, 't1pfa-cmap.ttx'),
    (['r', 'cs', '_1'], T1PFA_NAME, 't1pfa-cmap_cs1.ttx'),
    (['r', 'cl', '_2'], T1PFA_NAME, 't1pfa-cmap_cl2.ttx'),
    (['r'], UFO2_NAME, 'ufo2-cmap.ttx'),
    (['r', 'cs', '_4'], UFO2_NAME, 'ufo2-cmap_cs4.ttx'),
    (['r', 'cl', '_5'], UFO2_NAME, 'ufo2-cmap_cl5.ttx'),
    (['r'], UFO3_NAME, 'ufo3-cmap.ttx'),
    (['r', 'cs', '_4'], UFO3_NAME, 'ufo3-cmap_cs4.ttx'),
    (['r', 'cl', '_5'], UFO3_NAME, 'ufo3-cmap_cl5.ttx'),
    (['r'], CID_NAME, 'cidfont-cmap.ttx'),
    (['r', 'cs', '_2'], CID_NAME, 'cidfont-cmap_cs2.ttx'),
    (['r', 'cl', '_3'], CID_NAME, 'cidfont-cmap_cl3.ttx'),
])
def test_build_options_cs_cl_bug459(args, input_filename, ttx_filename):
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'o', f'_{actual_path}'] + args)
    actual_ttx = generate_ttx_dump(actual_path, ['cmap'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


@pytest.mark.parametrize('args, font, fontinfo', [
    (['cn'], 'type1', ''),
    (['cn'], 'no_notdef', ''),
    (['cn'], 'blank_notdef', ''),
    (['cn'], 'notdef_not1st', ''),
    (['cn', 'fi'], 'type1', 'fi'),
    (['cn', 'fi'], 'type1', 'fi2'),
    (['cn', 'fi'], 'no_notdef', 'fi'),
    (['cn', 'fi'], 'no_notdef', 'fi2'),
    (['cn', 'fi'], 'blank_notdef', 'fi'),
    (['cn', 'fi'], 'blank_notdef', 'fi2'),
    (['cn', 'fi'], 'notdef_not1st', 'fi'),
    (['cn', 'fi'], 'notdef_not1st', 'fi2'),
])
def test_cid_keyed_cff_bug470(args, font, fontinfo):
    if 'fi' in args:
        fontinfo_file = f'bug470/{fontinfo}.txt'
        args.append(f'_{get_input_path(fontinfo_file)}')
        ttx_file = f'bug470/{font}-fi.ttx'
    else:
        ttx_file = f'bug470/{font}.ttx'
    font_file = f'bug470/{font}.pfa'
    # 'dir=TEMP_DIR' is used for guaranteeing that the temp data is on same
    # file system as other data; if it's not, a file rename step made by
    # sfntedit will NOT work.
    actual_path = get_temp_file_path(directory=TEMP_DIR)
    runner(CMD + ['-o', 'f', f'_{get_input_path(font_file)}',
                        'o', f'_{actual_path}'] + args)
    actual_ttx = generate_ttx_dump(actual_path, ['CFF '])
    expected_ttx = get_expected_path(ttx_file)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


@pytest.mark.parametrize('opts', [
    [],
    ['r'],
    ['ga'],
    ['r', 'ga'],
    ['nga'],
    ['r', 'nga'],
    ['gf'],
    ['r', 'gf'],
    ['gf', 'ga'],
    ['gf', 'ga', 'nga'],  # GOADB will be applied, because 'ga' is first
    ['gf', 'nga'],
    ['gf', 'nga', 'ga'],  # GOADB not applied, because 'nga' is first
    ['r', 'ga', 'gf'],
    ['bit7n', 'bit8n', 'bit9n'],
    ['bit7y', 'bit8y', 'bit9y'],
    ['bit7y', 'bit8n', 'bit9n'],
    ['bit7y', 'bit8n', 'bit7n', 'bit8y'],
])
def test_GOADB_options_bug497(opts):
    font_path = get_input_path(T1PFA_NAME)
    feat_path = get_input_path('bug497/feat.fea')
    fmndb_path = get_input_path('bug497/fmndb.txt')
    goadb_path = get_input_path('bug497/goadb.txt')
    actual_path = get_temp_file_path()
    ttx_filename = '-'.join(['opts'] + opts) + '.ttx'

    args = []
    for opt in opts:  # order of the opts is important
        if opt == 'r':
            args.append(opt)
        elif opt == 'ga':
            args.append(opt)
        elif opt == 'nga':
            args.append(opt)
        elif opt == 'gf':
            args.extend([opt, f'_{goadb_path}'])
        elif 'bit' in opt:
            bit_num = int(opt[3])
            bit_bol = 'osbOn' if opt[4] == 'y' else 'osbOff'
            args.extend([bit_bol, f'_{bit_num}'])

    runner(CMD + ['-o', 'f', f'_{font_path}',
                        'o', f'_{actual_path}',
                        'ff', f'_{feat_path}',
                        'mf', f'_{fmndb_path}'] + args)
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(f'bug497/{ttx_filename}')
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])


@pytest.mark.parametrize('feat_name, has_warn', [('v0005', False),
                                                 ('ADBE', True)])
def test_fetch_font_version_bug610(feat_name, has_warn):
    input_filename = 't1pfa.pfa'
    feat_filename = f'bug610/{feat_name}.fea'
    otf_path = get_temp_file_path()

    stdout_path = runner(
        CMD + ['-s', '-o', 'r',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stdout_path, 'rb') as f:
        output = f.read()
    assert b"Revision 0.005" in output
    assert (b"[Warning] Major version number not in "
            b"range 1 .. 255" in output) is has_warn


def test_update_cff_bbox_bug617():
    input_filename = "bug617/font.pfa"
    goadb_filename = "bug617/goadb.txt"
    actual_path = get_temp_file_path()
    ttx_filename = "bug617.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'gf', f'_{get_input_path(goadb_filename)}',
                        'o', f'_{actual_path}', 'r', 'gs'])
    actual_ttx = generate_ttx_dump(actual_path, ['head', 'CFF '])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


@pytest.mark.parametrize('feat_filename', [
    'bug164/d1/d2/rel_to_main.fea',
    'bug164/d1/d2/rel_to_parent.fea',
])
def test_feature_includes_type1_bug164(feat_filename):
    input_filename = "bug164/d1/d2/font.pfa"
    otf_path = get_temp_file_path()

    runner(CMD + ['-o',
                  'f', f'_{get_input_path(input_filename)}',
                  'ff', f'_{get_input_path(feat_filename)}',
                  'o', f'_{otf_path}'])

    assert font_has_table(otf_path, 'head')


def test_feature_includes_ufo_bug164():
    input_filename = "bug164/d1/d2/font.ufo"
    otf_path = get_temp_file_path()

    runner(CMD + ['-o',
                  'f', f'_{get_input_path(input_filename)}',
                  'o', f'_{otf_path}'])

    assert font_has_table(otf_path, 'head')


def test_ttf_input_font_bug680():
    input_filename = 'bug680/font.ttf'
    feat_filename = 'bug680/features.fea'
    ttf_path = get_temp_file_path()

    runner(CMD + ['-o', 'r',
                  'f', f'_{get_input_path(input_filename)}',
                  'ff', f'_{get_input_path(feat_filename)}',
                  'o', f'_{ttf_path}'])

    for table_tag in ('head', 'hhea', 'maxp', 'OS/2', 'hmtx', 'cmap', 'fpgm',
                      'prep', 'cvt ', 'loca', 'glyf', 'name', 'post', 'gasp',
                      'BASE', 'GDEF', 'GPOS', 'GSUB'):
        assert font_has_table(ttf_path, table_tag)


def test_skip_ufo3_global_guides_bug700():
    input_filename = "bug700.ufo"
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'o', f'_{actual_path}'])
    assert font_has_table(actual_path, 'CFF ')


def test_outline_from_processed_layer_bug703():
    input_filename = 'bug703.ufo'
    ttx_filename = 'bug703.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['CFF '])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_unhandled_ufo_glif_token_bug705():
    input_filename = 'bug705.ufo'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o',
               'f', f'_{get_input_path(input_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert b"unhandled token: <foo" in output


def test_delete_zero_kb_font_on_fail_bug736():
    input_filename = 'bug736/font.pfa'
    feat_filename = 'bug736/feat.fea'
    out_filename = 'bug736/SourceSans-Test.otf'

    with pytest.raises(subprocess.CalledProcessError) as err:
        runner(CMD + ['-o',
                      'f', f'_{get_input_path(input_filename)}',
                      'ff', f'_{get_input_path(feat_filename)}'])
    assert err.value.returncode == 2
    assert os.path.exists(get_input_path(out_filename)) is False


def test_duplicate_warning_messages_bug751():
    input_filename = 'bug751.ufo'
    expected_path = get_expected_path('bug751.txt')
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o',
               'f', f'_{get_input_path(input_filename)}',
               'o', f'_{otf_path}'])

    assert differ([expected_path, stderr_path, '-l', '1',
                   '-s', 'Built development mode font'])


@pytest.mark.parametrize('feat_filename, rev_value, expected', [
    ('fontrev/fr_1_000.fea', '1', '1.001'),
    ('fontrev/fr_1_000.fea', '2.345', '2.345'),
    ('fontrev/fr_1_000.fea', '40000.000', '1.000')])
def test_update_font_revision(feat_filename, rev_value, expected):
    input_filename = "fontrev/font.ufo"
    otf_path = get_temp_file_path()
    tmp_fea_path = get_temp_file_path()
    copy2(get_input_path(feat_filename), tmp_fea_path)

    runner(CMD + ['-o',
                  'f', f'_{get_input_path(input_filename)}',
                  'ff', f'_{tmp_fea_path}',
                  'o', f'_{otf_path}',
                  'rev', f'_{rev_value}',
                  ])
    frstr = f'{get_font_revision(otf_path):.3f}'
    assert frstr == expected


def test_cli_numerics():
    input_filename = "clinum/font.otf"
    fontinfo_filename = "clinum/fontinfo"
    expected_msg_path = get_expected_path("clinum_output.txt")
    out_font_path = get_temp_file_path()
    stderr_path = runner(CMD + [
        '-s', '-e', '-o',
        'f', f'_{get_input_path(input_filename)}',
        'o', f'_{out_font_path}',
        'fi', f'_{get_input_path(fontinfo_filename)}'])
    assert differ([expected_msg_path, stderr_path,
                   '-r', r'^Built (development|release) mode font'])


@pytest.mark.parametrize('explicit_fmndb', [False, True])
def test_check_psname_in_fmndb_bug1171(explicit_fmndb):
    input_path = get_input_path('bug1171/font.ufo')
    fmndb_path = get_input_path('bug1171/FontMenuNameDB')
    expected_ttx = get_expected_path('bug1171.ttx')
    actual_path = get_temp_file_path()
    opts = ['-o', 'f', f'_{input_path}', 'o', f'_{actual_path}']
    if explicit_fmndb is True:
        opts.extend(['mf', f'_{fmndb_path}'])
    runner(CMD + opts)
    actual_ttx = generate_ttx_dump(actual_path, ['name'])
    assert differ([expected_ttx, actual_ttx,
                   '-s', '<ttFont sfntVersion',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])



import os
import pytest
from shutil import copy2, copytree
import subprocess

from fontTools.ttLib import TTFont
from afdko import ttfcomponentizer as ttfcomp

from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
)
from test_utils import (
    get_input_path,
    get_expected_path,
    generate_ttx_dump,
)
from runner import main as runner
from differ import main as differ

TOOL = 'ttfcomponentizer'
CMD = ['-t', TOOL]

TEST_TTF_FILENAME = 'ttfcomponentizer.ttf'
DESIGN_NAMES_LIST = ['acaron', 'acutecmb', 'caroncmb', 'design_name',
                     'gravecmb', 'tildecmb']
PRODCT_NAMES_LIST = ['production_name', 'uni01CE', 'uni0300', 'uni0301',
                     'uni0303', 'uni030C']
EMPTY_LIB = b"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN"
"http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
    <dict/>
</plist>
"""
EMPTY_PSKEY = b"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN"
"http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
    <dict>
    <key>public.postscriptNames</key>
    <dict/>
    </dict>
</plist>
"""


def _get_test_ttf_path():
    return get_input_path(TEST_TTF_FILENAME)


def _get_test_ufo_path():
    return _get_test_ttf_path()[:-3] + 'ufo'


def _read_file(file_path):
    with open(file_path, "rb") as f:
        return f.read()


def _write_file(file_path, data):
    with open(file_path, "wb") as f:
        f.write(data)


# -----
# Tests
# -----

def test_run_no_args():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main()
    assert exc_info.value.code == 2


def test_run_cli_no_args():
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        runner(CMD)
    assert exc_info.value.returncode == 2


def test_run_invalid_font_path():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main(['not_a_file'])
    assert exc_info.value.code == 2


def test_run_invalid_font_file():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main(['not_a_font.ttf'])
    assert exc_info.value.code == 2


def test_run_ufo_not_found():
    ttf_path = _get_test_ttf_path()
    temp_dir = get_temp_dir_path()
    save_path = get_temp_file_path(directory=temp_dir)
    copy2(ttf_path, save_path)
    assert ttfcomp.main([save_path]) == 1


def test_run_invalid_ufo():
    ttf_path = _get_test_ttf_path()
    temp_dir = get_temp_dir_path()
    save_path = get_temp_file_path(directory=temp_dir)
    ufo_path = save_path + '.ufo'
    copy2(ttf_path, save_path)
    copy2(ttf_path, ufo_path)
    assert ttfcomp.main([save_path]) == 1


def test_run_with_output_path():
    ttf_path = _get_test_ttf_path()
    save_path = get_temp_file_path()
    ttfcomp.main(['-o', save_path, ttf_path])
    gtable = TTFont(save_path)['glyf']
    composites = [gname for gname in gtable.glyphs if (
        gtable[gname].isComposite())]
    assert sorted(composites) == ['aa', 'aacute', 'uni01CE']


def test_run_cli_with_output_path():
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'o', f'_{actual_path}',
                  f'_{get_input_path(TEST_TTF_FILENAME)}'])
    actual_ttx = generate_ttx_dump(actual_path, ['maxp', 'glyf'])
    expected_ttx = get_expected_path('ttfcomponentizer.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_run_without_output_path():
    ttf_path = _get_test_ttf_path()
    ufo_path = _get_test_ufo_path()
    temp_dir = get_temp_dir_path()
    tmp_ufo_path = os.path.join(temp_dir, os.path.basename(ufo_path))
    save_path = get_temp_file_path(directory=temp_dir)
    copy2(ttf_path, save_path)
    copytree(ufo_path, tmp_ufo_path)
    ttfcomp.main([save_path])
    gtable = TTFont(save_path)['glyf']
    assert gtable['agrave'].isComposite() is False
    assert gtable['aacute'].isComposite() is True


def test_options_help():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main(['-h'])
    assert exc_info.value.code == 0


def test_options_version():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main(['--version'])
    assert exc_info.value.code == 0


def test_options_bogus_option():
    with pytest.raises(SystemExit) as exc_info:
        ttfcomp.main(['--bogus'])
    assert exc_info.value.code == 2


def test_get_ufo_path_found():
    path = _get_test_ttf_path()
    assert ttfcomp.get_ufo_path(path) == _get_test_ufo_path()


def test_get_ufo_path_not_found():
    # the UFO font is intentionally used as a test folder
    path = os.path.join(_get_test_ufo_path(), 'lib.plist')
    assert ttfcomp.get_ufo_path(path) is None


def test_get_goadb_path_found_same_folder():
    path = _get_test_ttf_path()
    assert os.path.basename(
        ttfcomp.get_goadb_path(path)) == ttfcomp.GOADB_FILENAME


def test_get_goadb_path_found_3_folders_up():
    # the UFO's glif file is intentionally used as a test file
    path = os.path.join(_get_test_ufo_path(), 'glyphs', 'a.glif')
    assert os.path.basename(
        ttfcomp.get_goadb_path(path)) == ttfcomp.GOADB_FILENAME


def test_get_goadb_path_not_found():
    path = os.path.dirname(_get_test_ttf_path())
    assert ttfcomp.get_goadb_path(path) is None


def test_get_goadb_names_mapping_goadb_not_found():
    path = os.path.dirname(_get_test_ttf_path())
    assert ttfcomp.get_goadb_names_mapping(path) == {}


def test_get_glyph_names_mapping_invalid_ufo():
    path = _get_test_ttf_path()
    assert ttfcomp.get_glyph_names_mapping(path) == (None, None)


def test_get_glyph_names_mapping_names_from_lib():
    result = ttfcomp.get_glyph_names_mapping(_get_test_ufo_path())
    assert sorted(result[1].keys()) == DESIGN_NAMES_LIST
    assert sorted(result[1].values()) == PRODCT_NAMES_LIST


def test_get_glyph_names_mapping_names_from_goadb():
    # the test UFO has a 'public.postscriptNames' in its lib;
    # empty the lib temporarily to force the reading of the GOADB
    ufo_path = _get_test_ufo_path()
    lib_path = os.path.join(ufo_path, 'lib.plist')
    lib_data = _read_file(lib_path)
    _write_file(lib_path, EMPTY_LIB)
    result = ttfcomp.get_glyph_names_mapping(ufo_path)
    _write_file(lib_path, lib_data)
    assert sorted(result[1].keys()) == DESIGN_NAMES_LIST
    assert sorted(result[1].values()) == PRODCT_NAMES_LIST


def test_warn_empty_psnames_key(capsys):
    # the test UFO has a 'public.postscriptNames' in its lib;
    # set the key to empty to ensure we get a warning
    ufo_path = _get_test_ufo_path()
    lib_path = os.path.join(ufo_path, 'lib.plist')
    lib_data = _read_file(lib_path)
    _write_file(lib_path, EMPTY_PSKEY)
    result = ttfcomp.get_glyph_names_mapping(ufo_path)
    _write_file(lib_path, lib_data)
    assert len(result[1]) == 0
    captured = capsys.readouterr()
    # check (partial) warning message in stdout:
    assert "WARNING: The contents of public.postscriptNames is empty." in captured.out  # noqa: E501


def test_componentize():
    ttf_path = _get_test_ttf_path()
    save_path = get_temp_file_path()
    ufo, ps_names = ttfcomp.get_glyph_names_mapping(_get_test_ufo_path())
    ttcomp_obj = ttfcomp.TTComponentizer(ufo, ps_names, ttf_path, save_path)
    ttcomp_obj.componentize()

    # 'get_composites_data' method
    comps_data = ttcomp_obj.composites_data
    comps_name_list = sorted(comps_data.keys())
    comps_comp_list = [comps_data[gname] for gname in comps_name_list]
    assert comps_name_list == ['aa', 'aacute', 'adieresis', 'atilde',
                               'uni01CE']
    assert comps_comp_list[1].names == ('a', 'uni0301')
    assert comps_comp_list[4].names == ('a', 'uni030C')
    assert comps_comp_list[1].positions == ((0, 0), (263.35, 0))
    assert comps_comp_list[4].positions == ((0, 0), (263, 0))

    # 'assemble_components' method
    comps_data = ttfcomp.ComponentsData()
    comps_data.names = ('a', 'uni01CE')
    comps_data.positions = ((0, 0), (263, 0))
    comps_data.same_advwidth = True
    comp_one, comp_two = ttcomp_obj.assemble_components(comps_data)
    assert comp_one.glyphName == 'a'
    assert comp_two.glyphName == 'uni01CE'
    assert (comp_one.x, comp_one.y) == (0, 0)
    assert (comp_two.x, comp_two.y) == (263, 0)
    assert comp_one.flags == 0x204
    assert comp_two.flags == 0x4



import argparse
import pytest
from os import linesep, path
import sys

from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
    get_resources_dir,
    get_font_format,
    validate_path,
    run_shell_command,
    run_shell_command_logging,
    get_shell_command_output,
    runShellCmd,
    runShellCmdLogging,
)
from test_utils import get_input_path

mac_linux_only = pytest.mark.skipif(
    sys.platform == 'win32', reason="Mac/Linux subprocess results")

win_only = pytest.mark.skipif(
    sys.platform != 'win32', reason="Windows subprocess results")


def test_validate_path_valid():
    assert path.exists(validate_path(get_temp_file_path())) is True


def test_validate_path_invalid():
    with pytest.raises(argparse.ArgumentTypeError):
        validate_path('not_a_file_or_folder')


@pytest.mark.parametrize('path_comp', [None, 'path_component'])
def test_get_temp_dir_path(path_comp):
    pth = get_temp_dir_path(path_comp)
    if path_comp:
        assert path.basename(pth) == path_comp
        assert path.isdir(path.dirname(pth))
    else:
        assert path.isdir(pth)


def test_get_resources_dir():
    assert path.isdir(get_resources_dir())


@pytest.mark.parametrize('filename, fontformat', [
    ('file.txt', None),
    ('font.ufo', 'UFO'),
    ('font.otf', 'OTF'),
    ('font.ttf', 'TTF'),
    ('font.ttc', 'TTC'),
    ('font.cff', 'CFF'),
    ('font.pfa', 'PFA'),
    ('font2.pfa', 'PFA'),
    ('font.pfb', 'PFB'),
    ('font.ps', 'PFC'),
])
def test_get_font_format(filename, fontformat):
    assert get_font_format(get_input_path(filename)) == fontformat


@pytest.mark.parametrize('cmd, rslt', [
    (['type1', '-h'], True),
    (['type1', 'foo'], False),
])
def test_run_shell_command(cmd, rslt):
    assert run_shell_command(cmd) is rslt


@pytest.mark.parametrize('cmd, rslt', [
    (['type1', get_input_path('type1.txt')], True),
    (['type1', get_input_path('type1.txt'), get_temp_file_path()], True),
    (['type1', 'foo'], False),
    ('type1 foo', False),
])
def test_run_shell_command_logging(cmd, rslt):
    assert run_shell_command_logging(cmd) is rslt


@pytest.mark.parametrize('cmd, rslt', [
    (['type1', '-h'], (True, f'usage: type1 [text [font]]{linesep}')),
    (['type1', 'foo'], (False, None)),  # goes thru the exception
])
def test_get_shell_command_output(cmd, rslt):
    assert get_shell_command_output(cmd) == rslt


def test_runShellCmd_timeout():
    assert runShellCmd('type1 -h', timeout=0) == ''


@mac_linux_only
@pytest.mark.parametrize('cmd, shell, str_out', [
    ('type1 -h', True, f'usage: type1 [text [font]]{linesep}'),
    ('type1 -h', False, ''),  # fails on Windows
    (['type1', '-h'], True, ''),  # fails on Windows
    (['type1', '-h'], False, f'usage: type1 [text [font]]{linesep}'),
])
def test_runShellCmd_mac_linux(cmd, shell, str_out):
    assert runShellCmd(cmd, shell=shell) == str_out


@win_only
@pytest.mark.parametrize('cmd, shell', [
    ('type1 -h', True),
    ('type1 -h', False),
    (['type1', '-h'], True),
    (['type1', '-h'], False),
])
def test_runShellCmd_windows(cmd, shell):
    # The shell command call always produces output on Windows,
    # regardless of the value of the shell (True/False) and
    # no matter if the command/args are a list or a string
    assert runShellCmd(cmd, shell=shell) == (
        f'usage: type1 [text [font]]{linesep}')


@pytest.mark.parametrize('cmd, shell, str_out', [
    ('type1 foo', True, ''),
    ('type1 foo', False, ''),
    (['type1', 'foo'], True, ''),
    (['type1', 'foo'], False, ''),
])
def test_runShellCmd_error(cmd, shell, str_out):
    assert runShellCmd(cmd, shell=shell) == str_out


@pytest.mark.parametrize('cmd, rtrn', [
    ('', 1),
    ('type1 foo', 1),
    ([''], 1),
    (['tx', '-h'], 0),
    (['tx', '-z'], 1),
    (['type1', get_input_path('type1.txt'), get_temp_file_path()], 0),
])
def test_runShellCmdLogging_shell_false(cmd, rtrn):
    assert runShellCmdLogging(cmd, False) == rtrn



import argparse
import os
import pytest
from shutil import rmtree, copytree

from afdko.makeinstancesufo import (
    main as mkinstufo,
    get_options,
    _split_comma_sequence,
    updateInstance,
)
from runner import main as runner
from differ import main as differ
from test_utils import get_input_path

TOOL = 'makeinstancesufo'

DATA_DIR = os.path.join(os.path.dirname(__file__), TOOL + '_data')
TEMP_DIR = os.path.join(DATA_DIR, "temp_output")


def _get_output_path(file_name, dir_name):
    return os.path.join(DATA_DIR, dir_name, file_name)


def setup_module():
    """
    Create the temporary output directory
    """
    rmtree(TEMP_DIR, ignore_errors=True)
    os.mkdir(TEMP_DIR)


def teardown_module():
    """
    teardown the temporary UFOs or the directory that holds them
    """
    rmtree(os.path.join(TEMP_DIR), True)
    rmtree(os.path.join(DATA_DIR, 'input', 'same_dir.ufo'), True)


# -----
# Tests
# -----

@pytest.mark.parametrize('args, ufo_filename', [
    (['_0'], 'extralight.ufo'),  # hint/remove overlap/normalize/round
    (['_1', 'a'], 'light.ufo'),  # no hint
    (['_2', 'r'], 'regular.ufo'),  # no round
    (['_3', 'r', 'n'], 'regular1.ufo'),  # no round & no normalize
    (['_4', 'c'], 'semibold.ufo'),  # no remove overlap
    (['_5', 'n'], 'bold.ufo'),  # no normalize
    (['_6', 'a', 'c', 'n'], 'black.ufo'),  # round only
    (['_7', 'a', 'c', 'n'], 'anisotropic.ufo'),
    (['_8', 'a', '=ufo-version', '_2'], 'ufo2.ufo'),  # no hint UFO v2
])
def test_options(args, ufo_filename):
    runner(['-t', TOOL, '-o', 'd',
            f'_{get_input_path("font.designspace")}', 'i'] + args)
    expected_path = _get_output_path(ufo_filename, 'expected_output')
    actual_path = _get_output_path(ufo_filename, 'temp_output')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('v_arg', ['', 'v', 'vv', 'vvv'])
def test_log_level(v_arg):
    dspath = get_input_path('font.designspace')
    v_val = len(v_arg)
    arg = [] if not v_val else [f'-{v_arg}']
    opts = get_options(['-d', dspath] + arg)
    assert opts.verbose == v_val


@pytest.mark.parametrize('str_seq, lst_seq', [
    ('10', [10]),
    ('11,13,19', [11, 13, 19]),
    ('16,0,13', [0, 13, 16]),
])
def test_split_comma_sequence(str_seq, lst_seq):
    assert _split_comma_sequence(str_seq) == lst_seq


@pytest.mark.parametrize('str_seq', ['0,a', '1,-2', '10-10', '5+4', '3*5'])
def test_split_comma_sequence_error(str_seq):
    with pytest.raises(argparse.ArgumentTypeError):
        _split_comma_sequence(str_seq)


@pytest.mark.parametrize('filename', [
    'noaxes', 'nosources', 'nosourcepath', 'badsourcepath',
    'noinstances', 'noinstancepath', ('badinstanceindex', 2)
])
def test_validate_designspace_doc_raise(filename):
    args = []
    if isinstance(filename, tuple):
        filename, idx = filename
        args = ['-i', f'{idx}']
    dspath = get_input_path(f'{filename}.designspace')
    assert mkinstufo(['-d', dspath] + args) == 1


@pytest.mark.parametrize('args', [
    [],  # checkoutlinesufo call
    ['-c'],  # psautohint call
])
def test_update_instance_raise(args):
    dspath = get_input_path('font.designspace')
    ufopath = get_input_path('invalid.ufo')
    opts = get_options(['-d', dspath] + args)
    with pytest.raises(SystemExit):
        updateInstance(ufopath, opts)


def test_filename_without_dir():
    instance_path = get_input_path('same_dir.ufo')
    assert not os.path.exists(instance_path)
    runner(['-t', TOOL, '-o', 'd',
            f'_{get_input_path("font.designspace")}', 'i', '_9'])
    assert os.path.exists(instance_path)


@pytest.mark.parametrize('args, ufo_filename', [
    (['_0'], 'ufo3regular.ufo'),  # hint/remove overlap/normalize/round
    (['_1'], 'ufo3regular.ufo'),  # for testing instance removal
    (['_2', 'a', 'c', 'n'], 'ufo3medium.ufo'),  # round only
    (['_3', 'a', '=ufo-version', '_2'], 'ufo3semibold.ufo'),  # no hint UFO v2
])
def test_ufo3_masters(args, ufo_filename):
    runner(['-t', TOOL, '-o', 'd',
            f'_{get_input_path("ufo3.designspace")}', 'i'] + args)
    expected_path = _get_output_path(ufo_filename, 'expected_output')
    actual_path = _get_output_path(ufo_filename, 'temp_output')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('filename', ['features_copy', 'features_nocopy'])
def test_features_copy(filename):
    # NOTE: This test was originally implemented without copying the expected
    # UFOs to a temp location, but the Windows64 build always modified the glif
    # files' line endings after the test was ran and this caused wheel problems
    # https://ci.appveyor.com/project/adobe-type-tools/afdko/builds/25459479
    # ---
    # First copy the expected UFOs into the temp folder. The UFOs need to
    # exist before makeinstancesufo is called because this test is about
    # checking that any existing features.fea files are preserved.
    paths = []
    for i in (1, 2):  # two instances
        ufo_filename = f'{filename}{i}.ufo'
        from_path = _get_output_path(ufo_filename, 'expected_output')
        to_path = os.path.join(TEMP_DIR, ufo_filename)
        copytree(from_path, to_path)
        paths.append((to_path, from_path))
    # run makeinstancesufo
    runner(['-t', TOOL, '-o', 'a', 'c', 'n', 'd',
            f'_{get_input_path(f"{filename}.designspace")}'])
    # assert the expected results
    for expected_path, actual_path in paths:
        assert differ([expected_path, actual_path])


@pytest.mark.parametrize('args, ufo_filename', [
    (['_0'], 'bend1.ufo'),
    (['_1'], 'bend2.ufo'),
    (['_2'], 'bend3.ufo'),
])
def test_bend_masters_mutator_math(args, ufo_filename):
    # MutatorMath 2.1.2 did not handle location bending properly which resulted
    # in incorrect interpolation outlines. This was fixed in 3.0.1.
    runner(['-t', TOOL, '-o', 'a', 'c', 'n', 'd',
            f'_{get_input_path("bend_test.designspace")}', 'i'] + args)
    expected_path = _get_output_path(ufo_filename, 'expected_output')
    actual_path = _get_output_path(ufo_filename, 'temp_output')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('args, ufo_filename', [
    (['_0'], 'bend1.ufo'),
    (['_1'], 'bend2.ufo'),
    (['_2'], 'bend3.ufo'),
])
def test_bend_masters_varlib(args, ufo_filename):
    # We should get the same output passing through varLib
    runner(['-t', TOOL, '-o', 'a', 'c', 'n', '=use-varlib', 'd',
            f'_{get_input_path("bend_test.designspace")}', 'i'] + args)
    expected_path = _get_output_path(ufo_filename, 'expected_output')
    actual_path = _get_output_path(ufo_filename, 'temp_output')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('use_varlib', [False, True])
def test_extrapolate(capfd, use_varlib):
    """
    Test extrapolating, with default (MutatorMath) and with varlib.
    Using varlib should fail (output should not be extrapolated) because
    extrapolation is not supported by varlib.
    """
    runner_args = ['-t', TOOL, '-o', 'v', 'd', f'_{get_input_path("extrapolation.designspace")}']  # noqa: E501
    if use_varlib:
        runner_args.append('=use-varlib')
    runner(runner_args)
    captured = capfd.readouterr()

    tool = "fontTools.varlib" if use_varlib else "MutatorMath"
    assert f"Building 2 instances with {tool}..." in captured.err

    for ufo_filename in ("Dummy-ExtraPlus.ufo", "Dummy-ExtraMinus.ufo"):
        expected_path = _get_output_path(ufo_filename, 'expected_output')
        actual_path = _get_output_path(ufo_filename, 'temp_output')

        if use_varlib:
            assert not differ([expected_path, actual_path])
            assert "Extrapolation is not supported with varlib (Dummy Extra Plus weight: 1500.0)" in captured.err  # noqa: E501
        else:
            assert differ([expected_path, actual_path])



from fontTools.ttLib import TTFont

from afdko.pdflib.fontpdf import (doTitle, FontPDFParams)
from afdko.pdflib.otfpdf import txPDFFont
from afdko.pdflib.pdfgen import Canvas

from test_utils import get_input_path

OTF_FONT = 'OTF.otf'


# -----
# Tests
# -----

def test_doTitle_pageIncludeTitle_1():
    with TTFont(get_input_path(OTF_FONT)) as otfont:
        params = FontPDFParams()
        assert params.pageIncludeTitle == 1
        pdfFont = txPDFFont(otfont, params)
        rt_canvas = Canvas("pdf_file_path")
        assert rt_canvas._code == []
        doTitle(rt_canvas, pdfFont, params, 1)
        assert len(rt_canvas._code)
        assert 'SourceSansPro-Black' in rt_canvas._code[1]


def test_doTitle_pageIncludeTitle_0():
    with TTFont(get_input_path(OTF_FONT)) as otfont:
        params = FontPDFParams()
        params.pageIncludeTitle = 0
        pdfFont = txPDFFont(otfont, params)
        rt_canvas = Canvas("pdf_file_path")
        assert rt_canvas._code == []
        doTitle(rt_canvas, pdfFont, params, 1)
        assert rt_canvas._code == []



import pytest
from shutil import copy2, copytree
import os

from booleanOperations.booleanGlyph import BooleanGlyph
from defcon import Glyph

from afdko.checkoutlinesufo import remove_tiny_sub_paths

from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
    get_font_format,
)
from test_utils import (
    get_input_path,
    get_expected_path,
    generate_ttx_dump,
)
from runner import main as runner
from differ import main as differ

TOOL = 'checkoutlinesufo'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

def test_remove_tiny_sub_paths_large_contour():
    g = Glyph()
    p = g.getPen()
    p.moveTo((100, 100))
    p.lineTo((200, 200))
    p.lineTo((0, 100))
    p.closePath()
    assert len(g[0]) == 3
    assert g.bounds == (0, 100, 200, 200)
    bg = BooleanGlyph(g)
    assert remove_tiny_sub_paths(bg, 25, []) == []


def test_remove_tiny_sub_paths_small_contour():
    g = Glyph()
    p = g.getPen()
    p.moveTo((1, 1))
    p.lineTo((2, 2))
    p.lineTo((0, 1))
    p.closePath()
    assert len(g[0]) == 3
    assert g.bounds == (0, 1, 2, 2)
    bg = BooleanGlyph(g)
    assert remove_tiny_sub_paths(bg, 25, []) == \
        ['Contour 0 is too small: bounding box is less than minimum area. '
         'Start point: ((1, 1)).']


@pytest.mark.parametrize('ufo_filename', ['ufo2.ufo', 'ufo3.ufo'])
@pytest.mark.parametrize('args, expct_label', [
    (['e', 'w', 'q'], 'dflt-layer.ufo'),
    (['e', 'q'], 'proc-layer.ufo'),
])
def test_remove_overlap_ufo(args, ufo_filename, expct_label):
    actual_path = get_temp_dir_path(ufo_filename)
    copytree(get_input_path(ufo_filename), actual_path)
    runner(CMD + ['-f', actual_path, '-o'] + args)
    expct_filename = f'{ufo_filename[:-4]}-{expct_label}'
    expected_path = get_expected_path(expct_filename)
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('filename, diffmode', [
    ('font.pfa', []),
    ('font.pfb', ['-m', 'bin']),
    ('font.cff', ['-m', 'bin']),
])
def test_remove_overlap_type1_cff(filename, diffmode):
    actual_path = get_temp_file_path()
    copy2(get_input_path(filename), actual_path)
    runner(CMD + ['-f', actual_path, '-o', 'e', 'q'])
    expected_path = get_expected_path(filename)
    assert differ([expected_path, actual_path] + diffmode)


def test_remove_overlap_otf():
    actual_path = get_temp_file_path()
    copy2(get_input_path('font.otf'), actual_path)
    runner(CMD + ['-f', actual_path, '-o', 'e', 'q'])
    actual_ttx = generate_ttx_dump(actual_path, ['CFF '])
    expected_ttx = get_expected_path('font.ttx')
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_bug790():
    """
    Test case where the result of overlap removal resulted in coincident points
    at contour start. Previously caused a crash when attempting to set start
    point on the second point.
    """
    ufoname = 'bug790.ufo'
    actual_path = get_temp_dir_path(ufoname)
    copytree(get_input_path(ufoname), actual_path)
    runner(CMD + ['-f', actual_path, '-o', 'e'])
    expected_path = get_expected_path(ufoname)
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('filename, diffmode', [
    ('cidfont.subset', ['-m', 'bin'])
])
def test_cidkeyed_remove_overlap(filename, diffmode):
    actual_path = get_temp_file_path()
    copy2(get_input_path(filename), actual_path)
    runner(CMD + ['-f', actual_path, '-o', 'e', 'q', '=no-overlap-checks'])
    expected_path = get_expected_path('cidfont.subset.checked')
    assert differ([expected_path, actual_path] + diffmode)


@pytest.mark.parametrize('input_font, expected_font', [
    ('ufo3.ufo', 'ufo3-proc-layer.ufo'),
    ('font.pfa', 'font.pfa'),
])
def test_output_file_option(input_font, expected_font):
    """
    Test the '-o' (output file) option.
    """
    in_path = get_input_path(input_font)
    out_path = os.path.join(get_temp_dir_path(), input_font)
    expected_path = get_expected_path(expected_font)
    runner(CMD + ['-f', in_path, '-o', 'e', 'o', '_' + out_path])

    assert get_font_format(out_path) == get_font_format(in_path)
    assert differ([expected_path, out_path])



# Copyright 2017 Adobe. All rights reserved.

import ast
import hashlib
import os
import re
from collections import OrderedDict

from fontPens.thresholdPen import ThresholdPen
from fontTools.misc import etree as ET
from fontTools.misc import plistlib
from fontTools.pens.recordingPen import RecordingPen
from fontTools.ufoLib import UFOReader

from psautohint.ufoFont import norm_float

from afdko import fdkutils

__version__ = '1.35.2'

__doc__ = """
ufotools.py v1.35.2 Jul 30 2020

Originally developed to work with 'bez' files and UFO fonts in support of
the autohint tool, ufotools.py is now only used in checkoutlinesufo (since
autohint has been dropped in favor of psautohint). Some references to "bez"
and "autohint" remain in the tool its documentation.

Users should *NOT* rely on long-term support of methods or classes within
this module directly. It is intended for AFDKO-internal use and not as a
general-purpose library outside of AFDKO.

The basic model is:
 - read GLIF file
 - transform GLIF XML element to bez file
 - call FDK tool on bez file
 - transform new bez file to GLIF XML element with new data, and save in list

After all glyphs are done save all the new GLIF XML elements to GLIF
files, and update the hash map.

A complication in the Adobe UFO workflow comes from the fact we want to
make sure that checkoutlines and autohint have been run on each glyph in
a UFO font, when building an OTF font from the UFO font. We need to run
checkoutlines, because we no longer remove overlaps from source UFO font
data, because this can make revising a glyph much easier. We need to run
autohint, because the glyphs must be hinted after checkoutlines is run,
and in any case we want all glyphs to have been autohinted. The problem
with this is that it can take a minute or two to run autohint or
checkoutlines on a 2K-glyph font. The way we avoid this is to make and
keep a hash of the source glyph drawing operators for each tool. When
the tool is run, it calculates a hash of the source glyph, and compares
it to the saved hash. If these are the same, the tool can skip the
glyph. This saves a lot of time: if checkoutlines and autohint are run
on all glyphs in a font, then a second pass is under 2 seconds.

Another issue is that since we no longer remove overlaps from the source
glyph files, checkoutlines must write any edited glyph data to a
different layer in order to not destroy the source data. The ufotools
defines an Adobe-specific glyph layer for processed glyphs, named
"glyphs.com.adobe.type.processedGlyphs".
checkoutlines writes new glyph files to the processed glyphs layer only
when it makes a change to the glyph data.

When the autohint program is run, the ufotools must be able to tell
whether checkoutlines has been run and has altered a glyph: if so, the
input file needs to be from the processed glyph layer, else it needs to
be from the default glyph layer.

The way the hashmap works is that we keep an entry for every glyph that
has been processed, identified by a hash of its marking path data. Each
entry contains:
- a hash of the glyph point coordinates, from the default layer.
This is set after a program has been run.
- a history list: a list of the names of each program that has been run,
  in order.
- an editStatus flag.
Altered GLIF data is always written to the Adobe processed glyph layer. The
program may or may not have altered the outline data. For example, autohint
adds private hint data, and adds names to points, but does not change any
paths.

If the stored hash for the glyph does not exist, the ufotools lib will save the
new hash in the hash map entry and will set the history list to contain just
the current program. The program will read the glyph from the default layer.

If the stored hash matches the hash for the current glyph file in the default
layer, and the current program name is in the history list,then ufotools
will return "skip=1", and the calling program may skip the glyph.

If the stored hash matches the hash for the current glyph file in the default
layer, and the current program name is not in the history list, then the
ufotools will return "skip=0". If the font object field 'usedProcessedLayer' is
set True, the program will read the glyph from the from the Adobe processed
layer, if it exists, else it will always read from the default layer.

If the hash differs between the hash map entry and the current glyph in the
default layer, and usedProcessedLayer is False, then ufotools will return
"skip=0". If usedProcessedLayer is True, then the program will consult the list
of required programs. If any of these are in the history list, then the program
will report an error and return skip =1, else it will return skip=1. The
program will then save the new hash in the hash map entry and reset the history
list to contain just the current program. If the old and new hash match, but
the program name is not in the history list, then the ufotools will not skip
the glyph, and will add the program name to the history list.


The only tools using this are, at the moment, checkoutlines, checkoutlinesufo
and autohint. checkoutlines and checkoutlinesufo use the hash map to skip
processing only when being used to edit glyphs, not when reporting them.
checkoutlines necessarily flattens any components in the source glyph file to
actual outlines. autohint adds point names, and saves the hint data as a
private data in the new GLIF file.

autohint saves the hint data in the GLIF private data area, /lib/dict,
as a key and data pair. See below for the format.

autohint started with _hintFormat1_, a reasonably compact XML representation of
the data. In Sep 2105, autohhint switched to _hintFormat2_ in order to be plist
compatible. This was necessary in order to be compatible with the UFO spec, by
was driven more immediately by the fact the the UFO font file normalization
tools stripped out the _hintFormat1_ hint data as invalid elements.
"""

_hintFormat1_ = """

Deprecated. See _hintFormat2_ below.

A <hintset> element identifies a specific point by its name, and
describes a new set of stem hints which should be applied before the
specific point.

A <flex> element identifies a specific point by its name. The point is
the first point of a curve. The presence of the <flex> element is a
processing suggestion, that the curve and its successor curve should
be converted to a flex operator.

One challenge in applying the hintset and flex elements is that in the
GLIF format, there is no explicit start and end operator: the first path
operator is both the end and the start of the path. I have chosen to
convert this to T1 by taking the first path operator, and making it a
move-to. I then also use it as the last path operator. An exception is a
line-to; in T1, this is omitted, as it is implied by the need to close
the path. Hence, if a hintset references the first operator, there is a
potential ambiguity: should it be applied before the T1 move-to, or
before the final T1 path operator? The logic here applies it before the
move-to only.

 <glyph>
...
    <lib>
        <dict>
            <key><com.adobe.type.autohint><key>
            <data>
                <hintSetList>
                    <hintset pointTag="point name">
                      (<hstem pos="<decimal value>" width="decimal value" />)*
                      (<vstem pos="<decimal value>" width="decimal value" />)*
                      <!-- where n1-5 are decimal values -->
                      <hstem3 stem3List="n0,n1,n2,n3,n4,n5" />*
                      <!-- where n1-5 are decimal values -->
                      <vstem3 stem3List="n0,n1,n2,n3,n4,n5" />*
                    </hintset>*
                    (<hintSetList>*
                        (<hintset pointIndex="positive integer">
                            (<stemindex>positive integer</stemindex>)+
                        </hintset>)+
                    </hintSetList>)*
                    <flexList>
                        <flex pointTag="point Name" />
                    </flexList>*
                </hintSetList>
            </data>
        </dict>
    </lib>
</glyph>

Example from "B" in SourceCodePro-Regular
<key><com.adobe.type.autohint><key>
<data>
<hintSetList id="64bf4987f05ced2a50195f971cd924984047eb1d79c8c43e6a0054f59cc85
dea23a49deb20946a4ea84840534363f7a13cca31a81b1e7e33c832185173369086">
    <hintset pointTag="hintSet0000">
        <hstem pos="0" width="28" />
        <hstem pos="338" width="28" />
        <hstem pos="632" width="28" />
        <vstem pos="100" width="32" />
        <vstem pos="496" width="32" />
    </hintset>
    <hintset pointTag="hintSet0005">
        <hstem pos="0" width="28" />
        <hstem pos="338" width="28" />
        <hstem pos="632" width="28" />
        <vstem pos="100" width="32" />
        <vstem pos="454" width="32" />
        <vstem pos="496" width="32" />
    </hintset>
    <hintset pointTag="hintSet0016">
        <hstem pos="0" width="28" />
        <hstem pos="338" width="28" />
        <hstem pos="632" width="28" />
        <vstem pos="100" width="32" />
        <vstem pos="496" width="32" />
    </hintset>
</hintSetList>
</data>

"""

_hintFormat2_ = """

A <dict> element in the hintSetList array identifies a specific point by its
name, and describes a new set of stem hints which should be applied before the
specific point.

A <string> element in the flexList identifies a specific point by its name.
The point is the first point of a curve. The presence of the element is a
processing suggestion, that the curve and its successor curve should be
converted to a flex operator.

One challenge in applying the hintSetList and flexList elements is that in
the GLIF format, there is no explicit start and end operator: the first path
operator is both the end and the start of the path. I have chosen to convert
this to T1 by taking the first path operator, and making it a move-to. I then
also use it as the last path operator. An exception is a line-to; in T1, this
is omitted, as it is implied by the need to close the path. Hence, if a hintset
references the first operator, there is a potential ambiguity: should it be
applied before the T1 move-to, or before the final T1 path operator? The logic
here applies it before the move-to only.
 <glyph>
...
    <lib>
        <dict>
            <key><com.adobe.type.autohint></key>
            <dict>
                <key>id</key>
                <string> <fingerprint for glyph> </string>
                <key>hintSetList</key>
                <array>
                    <dict>
                      <key>pointTag</key>
                      <string> <point name> </string>
                      <key>stems</key>
                      <array>
                        <string>hstem <position value> <width value></string>*
                        <string>vstem <position value> <width value></string>*
                        <string>hstem3 <position value 0>...<position value 5>
                        </string>*
                        <string>vstem3 <position value 0>...<position value 5>
                        </string>*
                      </array>
                    </dict>*
                </array>

                <key>flexList</key>*
                <array>
                    <string><point name></string>+
                </array>
            </dict>
        </dict>
    </lib>
</glyph>

Example from "B" in SourceCodePro-Regular
<key><com.adobe.type.autohint><key>
<dict>
    <key>id</key>
    <string>64bf4987f05ced2a50195f971cd924984047eb1d79c8c43e6a0054f59cc85dea23
    a49deb20946a4ea84840534363f7a13cca31a81b1e7e33c832185173369086</string>
    <key>hintSetList</key>
    <array>
        <dict>
            <key>pointTag</key>
            <string>hintSet0000</string>
            <key>stems</key>
            <array>
                <string>hstem 338 28</string>
                <string>hstem 632 28</string>
                <string>hstem 100 32</string>
                <string>hstem 496 32</string>
            </array>
        </dict>
        <dict>
            <key>pointTag</key>
            <string>hintSet0005</string>
            <key>stems</key>
            <array>
                <string>hstem 0 28</string>
                <string>hstem 338 28</string>
                <string>hstem 632 28</string>
                <string>hstem 100 32</string>
                <string>hstem 454 32</string>
                <string>hstem 496 32</string>
            </array>
        </dict>
        <dict>
            <key>pointTag</key>
            <string>hintSet0016</string>
            <key>stems</key>
            <array>
                <string>hstem 0 28</string>
                <string>hstem 338 28</string>
                <string>hstem 632 28</string>
                <string>hstem 100 32</string>
                <string>hstem 496 32</string>
            </array>
        </dict>
    </array>
<dict>

"""

# UFO names
kDefaultGlyphsLayerName = "public.default"
kDefaultGlyphsLayer = "glyphs"
kProcessedGlyphsLayerName = "com.adobe.type.processedglyphs"
kProcessedGlyphsLayer = "glyphs.%s" % kProcessedGlyphsLayerName
DEFAULT_LAYER_ENTRY = [kDefaultGlyphsLayerName, kDefaultGlyphsLayer]
PROCESSED_LAYER_ENTRY = [kProcessedGlyphsLayerName, kProcessedGlyphsLayer]

kFontInfoName = "fontinfo.plist"
kContentsName = "contents.plist"
kLibName = "lib.plist"
kPublicGlyphOrderKey = "public.glyphOrder"

kAdobeDomainPrefix = "com.adobe.type"
kAdobHashMapName = "%s.processedHashMap" % kAdobeDomainPrefix
kAdobHashMapVersionName = "hashMapVersion"
kAdobHashMapVersion = (1, 0)  # If major version differs, do not use.
kAutohintName = "autohint"
kCheckOutlineName = "checkOutlines"
kCheckOutlineNameUFO = "checkOutlines"

kOutlinePattern = re.compile(r"<outline.+?outline>", re.DOTALL)

COMP_TRANSFORM = OrderedDict([
    ('xScale', '1'),
    ('xyScale', '0'),
    ('yxScale', '0'),
    ('yScale', '1'),
    ('xOffset', '0'),
    ('yOffset', '0')
])


class UFOParseError(Exception):
    pass


class UFOFontData(object):
    def __init__(self, parentPath, useHashMap, programName):
        self.parentPath = parentPath
        self.glyphMap = {}
        self.processedLayerGlyphMap = {}
        self.newGlyphMap = {}
        self.glyphList = []
        self.fontInfo = None
        # If False, will skip reading hashmap and
        # testing to see if glyph can be skipped.
        # Should be used only when calling program is
        # running in report mode only, and not changing
        # any glyph data.
        self.useHashMap = useHashMap
        # Used to skip getting glyph data when glyph
        # hash matches hash of current glyph data.
        self.hashMap = {}
        self.fontDict = None
        self.programName = programName
        self.curSrcDir = None
        self.hashMapChanged = False
        self.glyphDefaultDir = os.path.join(self.parentPath, "glyphs")
        self.glyphLayerDir = os.path.join(self.parentPath,
                                          kProcessedGlyphsLayer)
        self.glyphWriteDir = self.glyphLayerDir
        self.historyList = []
        self.requiredHistory = []  # See documentation above.
        # If False, then read data only from the default layer;
        # else read glyphs from processed layer, if it exists.
        self.useProcessedLayer = False
        # If True, then write data to the default layer
        self.writeToDefaultLayer = False
        # If True, then do not skip any glyphs.
        self.doAll = False
        # track whether checkSkipGLyph has deleted an
        # out-of-date glyph from the processed glyph layer
        self.deletedGlyph = False
        # If true, do NOT round x,y values when processing
        self.allowDecimalCoords = False

        self.glyphSet = UFOReader(self.parentPath,
                                  validate=False).getGlyphSet(None)

    def getGlyphMap(self):
        if len(self.glyphMap) == 0:
            self.loadGlyphMap()
        return self.glyphMap

    def readHashMap(self):
        hashPath = os.path.join(self.parentPath, "data", kAdobHashMapName)
        if os.path.exists(hashPath):
            with open(hashPath, "r", encoding='utf-8') as fp:
                data = fp.read()
            newMap = ast.literal_eval(data)
        else:
            newMap = {kAdobHashMapVersionName: kAdobHashMapVersion}

        try:
            version = newMap[kAdobHashMapVersionName]
            if version[0] > kAdobHashMapVersion[0]:
                raise UFOParseError("Hash map version is newer than program. "
                                    "Please update the FDK")
            elif version[0] < kAdobHashMapVersion[0]:
                print("Updating hash map: was older version")
                newMap = {kAdobHashMapVersionName: kAdobHashMapVersion}
        except KeyError:
            print("Updating hash map: was older version")
            newMap = {kAdobHashMapVersionName: kAdobHashMapVersion}
        self.hashMap = newMap

    def writeHashMap(self):
        hashMap = self.hashMap
        if len(hashMap) == 0:
            return  # no glyphs were processed.

        hashDir = os.path.join(self.parentPath, "data")
        if not os.path.exists(hashDir):
            os.makedirs(hashDir)
        hashPath = os.path.join(hashDir, kAdobHashMapName)

        hasMapKeys = sorted(hashMap.keys())
        data = ["{"]
        for gName in hasMapKeys:
            data.append("'%s': %s," % (gName, hashMap[gName]))
        data.append("}")
        data.append("")
        data = '\n'.join(data)
        with open(hashPath, "w") as fp:
            fp.write(data)

    def getGlyphSrcPath(self, glyphName):
        if len(self.glyphMap) == 0:
            self.loadGlyphMap()
        glyphFileName = self.glyphMap[glyphName]

        # Try for processed layer first
        if self.useProcessedLayer and self.processedLayerGlyphMap:
            try:
                glyphFileName = self.processedLayerGlyphMap[glyphName]
                self.curSrcDir = self.glyphLayerDir
                glyphPath = os.path.join(self.glyphLayerDir, glyphFileName)
                if os.path.exists(glyphPath):
                    return glyphPath
            except KeyError:
                pass

        self.curSrcDir = self.glyphDefaultDir
        glyphPath = os.path.join(self.curSrcDir, glyphFileName)

        return glyphPath

    def getGlyphDefaultPath(self, glyphName):
        if len(self.glyphMap) == 0:
            self.loadGlyphMap()
        glyphFileName = self.glyphMap[glyphName]
        glyphPath = os.path.join(self.glyphDefaultDir, glyphFileName)
        return glyphPath

    def getGlyphProcessedPath(self, glyphName):
        if len(self.glyphMap) == 0:
            self.loadGlyphMap()
        if not self.processedLayerGlyphMap:
            return None

        try:
            glyphFileName = self.processedLayerGlyphMap[glyphName]
            glyphPath = os.path.join(self.glyphLayerDir, glyphFileName)
        except KeyError:
            glyphPath = None
        return glyphPath

    def checkSkipGlyph(self, glyphName, newSrcHash, doAll):
        skip = False
        if not self.useHashMap:
            return skip

        if len(self.hashMap) == 0:
            # Hash maps have not yet been read in. Get them.
            self.readHashMap()

        hashEntry = srcHash = None
        historyList = []
        programHistoryIndex = -1  # not found in historyList

        # Get hash entry for glyph
        try:
            hashEntry = self.hashMap[glyphName]
            srcHash, historyList = hashEntry
            try:
                programHistoryIndex = historyList.index(self.programName)
            except ValueError:
                pass
        except KeyError:
            # Glyph is as yet untouched by any program.
            pass

        if (srcHash == newSrcHash):
            if (programHistoryIndex >= 0):
                # The glyph has already been processed by this program,
                # and there have been no changes since.
                skip = True and (not doAll)
            if not skip:
                # case for Checkoutlines
                if not self.useProcessedLayer:
                    self.hashMapChanged = True
                    self.hashMap[glyphName] = [newSrcHash, [self.programName]]
                    glyphPath = self.getGlyphProcessedPath(glyphName)
                    if glyphPath and os.path.exists(glyphPath):
                        os.remove(glyphPath)
                else:
                    if (programHistoryIndex < 0):
                        historyList.append(self.programName)
        else:
            # case for autohint
            if self.useProcessedLayer:
                # Default layer glyph and stored glyph hash differ, and
                # useProcessedLayer is True. If any of the programs in
                # requiredHistory in are in the historyList, we need to
                # complain and skip.
                foundMatch = False
                if len(historyList) > 0:
                    for programName in self.requiredHistory:
                        if programName in historyList:
                            foundMatch = True
                if foundMatch:
                    print("Error. Glyph '%s' has been edited. You must first "
                          "run '%s' before running '%s'. Skipping." %
                          (glyphName, self.requiredHistory, self.programName))
                    skip = True

            # If the source hash has changed, we need to
            # delete the processed layer glyph.
            self.hashMapChanged = True
            self.hashMap[glyphName] = [newSrcHash, [self.programName]]
            glyphPath = self.getGlyphProcessedPath(glyphName)
            if glyphPath and os.path.exists(glyphPath):
                os.remove(glyphPath)
                self.deletedGlyph = True

        return skip

    @staticmethod
    def getGlyphXML(glyphDir, glyphFileName):
        glyphPath = os.path.join(glyphDir, glyphFileName)  # default
        etRoot = ET.ElementTree()
        glifXML = etRoot.parse(glyphPath)
        outlineXML = glifXML.find("outline")
        try:
            widthXML = glifXML.find("advance")
            if widthXML is not None:
                width = round(ast.literal_eval(widthXML.get("width", '0')), 9)
            else:
                width = 0
        except UFOParseError as e:
            print(f"Error. skipping glyph '{glyphFileName}' because "
                  f"of parse error: {str(e)}")
            return None, None, None
        return width, glifXML, outlineXML

    def getOrSkipGlyph(self, glyphName, doAll=0):
        # Get default glyph layer data, so we can check if the glyph
        # has been edited since this program was last run.
        # If the program name is in the history list, and the srcHash
        # matches the default glyph layer data, we can skip.
        if len(self.glyphMap) == 0:
            self.loadGlyphMap()
        glyphFileName = self.glyphMap.get(glyphName)
        if not glyphFileName:
            return None, True  # skip

        width, glifXML, outlineXML = self.getGlyphXML(self.glyphDefaultDir,
                                                      glyphFileName)
        if glifXML is None:
            return None, True  # skip

        # Hash is always from the default glyph layer.
        useDefaultGlyphDir = True
        newHash, _ = self.buildGlyphHashValue(
            width, outlineXML, glyphName, useDefaultGlyphDir)
        skip = self.checkSkipGlyph(glyphName, newHash, doAll)

        # If self.useProcessedLayer and there is a glyph in the
        # processed layer, get the outline from that.
        if self.useProcessedLayer and self.processedLayerGlyphMap:
            try:
                glyphFileName = self.processedLayerGlyphMap[glyphName]
            except KeyError:
                pass
            glyphPath = os.path.join(self.glyphLayerDir, glyphFileName)
            if os.path.exists(glyphPath):
                width, glifXML, _ = self.getGlyphXML(
                    self.glyphLayerDir, glyphFileName)
                if glifXML is None:
                    return None, True  # skip

        return width, skip

    def loadGlyphMap(self):
        # Need to both get the list of glyphs from contents.plist, and also
        # the glyph order. The latter is take from the public.glyphOrder key
        # in lib.plist, if it exists, else it is taken from the contents.plist
        # file. Any glyphs in contents.plist which are not named in the
        # public.glyphOrder are sorted after all glyphs which are named in the
        # public.glyphOrder,, in the order that they occured in contents.plist.
        contentsPath = os.path.join(self.parentPath, "glyphs", kContentsName)
        self.glyphMap, self.glyphList = parsePList(contentsPath)
        orderPath = os.path.join(self.parentPath, kLibName)
        self.orderMap = parseGlyphOrder(orderPath)
        if self.orderMap is not None:
            orderIndex = len(self.orderMap)
            orderList = []

            # If there are glyphs in the font which are not named in the
            # public.glyphOrder entry, add then in the order of the
            # contents.plist file.
            for glyphName in self.glyphList:
                try:
                    entry = [self.orderMap[glyphName], glyphName]
                except KeyError:
                    entry = [orderIndex, glyphName]
                    self.orderMap[glyphName] = orderIndex
                    orderIndex += 1
                orderList.append(entry)
            orderList.sort()
            self.glyphList = []
            for entry in orderList:
                self.glyphList.append(entry[1])
        else:
            self.orderMap = {}
            numGlyphs = len(self.glyphList)
            for i in range(numGlyphs):
                self.orderMap[self.glyphList[i]] = i

        # We also need to get the glyph map for the processed layer,
        # and use this when the glyph is read from the processed layer.
        # Because checkoutlinesufo used the defcon library, it can write
        # glyph file names that differ from what is in the default glyph layer.
        contentsPath = os.path.join(self.glyphLayerDir, kContentsName)
        if os.path.exists(contentsPath):
            self.processedLayerGlyphMap, self.processedLayerGlyphList = \
                parsePList(contentsPath)

    @staticmethod
    def _rd_val(str_val):
        """Round and normalize a (string) GLIF value"""
        return repr(norm_float(round(ast.literal_eval(str_val), 9)))

    def buildGlyphHashValue(self, width, outlineXML, glyphName,
                            useDefaultGlyphDir, level=0):
        """
        glyphData must be the official <outline> XML from a GLIF.
        We skip contours with only one point.
        """
        dataList = ["w%s" % norm_float(round(width, 9))]
        if level > 10:
            raise UFOParseError(
                "In parsing component, exceeded 10 levels of reference. "
                "'%s'. " % (glyphName))
        # <outline> tag is optional per spec., e.g. space glyph
        # does not necessarily have it.
        if outlineXML is not None:
            for childContour in outlineXML:
                if childContour.tag == "contour":
                    if len(childContour) < 2:
                        continue
                    for child in childContour:
                        if child.tag == "point":
                            ptType = child.get("type")
                            pointType = '' if ptType is None else ptType[0]
                            x = self._rd_val(child.get("x"))
                            y = self._rd_val(child.get("y"))
                            dataList.append("%s%s%s" % (pointType, x, y))

                elif childContour.tag == "component":
                    # append the component hash.
                    compGlyphName = childContour.get("base")

                    if compGlyphName is None:
                        raise UFOParseError(
                            "'%s' is missing the 'base' attribute in a "
                            "component." % glyphName)

                    dataList.append("%s%s" % ("base:", compGlyphName))

                    if useDefaultGlyphDir:
                        try:
                            componentPath = self.getGlyphDefaultPath(
                                compGlyphName)
                        except KeyError:
                            raise UFOParseError(
                                "'%s' component glyph is missing from "
                                "contents.plist." % (compGlyphName))
                    else:
                        # If we are not necessarily using the default layer
                        # for the main glyph, then a missing component may not
                        # have been processed, and may just be in the default
                        # layer. We need to look for component glyphs in the
                        # src list first, then in the defualt layer.
                        try:
                            componentPath = self.getGlyphSrcPath(compGlyphName)
                            if not os.path.exists(componentPath):
                                componentPath = self.getGlyphDefaultPath(
                                    compGlyphName)
                        except KeyError:
                            try:
                                componentPath = self.getGlyphDefaultPath(
                                    compGlyphName)
                            except KeyError:
                                raise UFOParseError(
                                    "'%s' component glyph is missing from "
                                    "contents.plist." % (compGlyphName))

                    if not os.path.exists(componentPath):
                        raise UFOParseError(
                            "'%s' component file is missing: '%s'." %
                            (compGlyphName, componentPath))

                    etRoot = ET.ElementTree()

                    # Collect transform values
                    for trans_key, flbk_val in COMP_TRANSFORM.items():
                        value = childContour.get(trans_key, flbk_val)
                        dataList.append(self._rd_val(value))

                    componentXML = etRoot.parse(componentPath)
                    componentOutlineXML = componentXML.find("outline")
                    _, componentDataList = self.buildGlyphHashValue(
                        width, componentOutlineXML, glyphName,
                        useDefaultGlyphDir, level + 1)
                    dataList.extend(componentDataList)
        data = "".join(dataList)
        if len(data) >= 128:
            data = hashlib.sha512(data.encode("ascii")).hexdigest()
        return data, dataList

    def close(self):
        if self.hashMapChanged:
            self.writeHashMap()
            self.hashMapChanged = False


def parseGlyphOrder(filePath):
    orderMap = None
    if os.path.exists(filePath):
        publicOrderDict, _ = parsePList(filePath, kPublicGlyphOrderKey)
        if publicOrderDict is not None:
            orderMap = {}
            glyphList = publicOrderDict[kPublicGlyphOrderKey]
            numGlyphs = len(glyphList)
            for i in range(numGlyphs):
                orderMap[glyphList[i]] = i
    return orderMap


def parsePList(filePath, dictKey=None):
    # If dictKey is defined, parse and return only the data for that key.
    #
    # Updates July 2019:
    #  - use fontTools.misc.plistlib instead of ET to parse
    #  - use built-in OrderedDict as the dict_type to preserve ordering
    #  - use simpler filtering for non-None dictKey

    plistDict = {}
    plistKeys = []

    with open(filePath, 'r', encoding='utf-8') as fp:
        plistDict = plistlib.load(fp, dict_type=OrderedDict)

    if dictKey is not None:
        if dictKey in plistDict:
            plistDict = {dictKey: plistDict[dictKey]}
        else:
            plistDict = None

    if plistDict is not None:
        plistKeys = list(plistDict.keys())

    return plistDict, plistKeys


def addWhiteSpace(parent, level):
    child = None
    childIndent = '\n' + ("  " * (level + 1))
    prentIndent = '\n' + ("  " * (level))
    # print("parent Tag", parent.tag, repr(parent.text), repr(parent.tail))
    for child in parent:
        child.tail = childIndent
        addWhiteSpace(child, level + 1)
    if child is not None:
        if parent.text is None:
            parent.text = childIndent
        child.tail = prentIndent
        # print("lastChild Tag", child.tag, repr(child.text),
        #       repr(child.tail), "parent Tag", parent.tag)


def regenerate_glyph_hashes(ufo_font_data):
    """
    The handling of the glyph hashes is super convoluted.
    This method fixes https://github.com/adobe-type-tools/afdko/issues/349
    """
    for gname, gfilename in ufo_font_data.getGlyphMap().items():
        gwidth, _, outline_xml = ufo_font_data.getGlyphXML(
            ufo_font_data.glyphDefaultDir, gfilename)
        hash_entry = ufo_font_data.hashMap.get(gname, None)
        if not hash_entry:
            continue
        ghash, _ = ufo_font_data.buildGlyphHashValue(
            gwidth, outline_xml, gname, True)
        hash_entry[0] = ghash


def checkHashMaps(fontPath, doSync):
    """
    Checks if the hashes of the glyphs in the default layer match the hash
    values stored in the UFO's 'data/com.adobe.type.processedHashMap' file.

    Returns a tuple of a boolean and a list. The boolean is True if all glyph
    hashes matched. The list contains strings that report the glyph names
    whose hash did not match.

    If doSync is True, it will delete any glyph in the processed glyph
    layer directory which does not have a matching glyph in the default
    layer, or whose source glyph hash does not match. It will then update
    the contents.plist file for the processed glyph layer, and delete
    the program specific hash maps.
    """
    msgList = []
    allMatch = True

    ufoFontData = UFOFontData(fontPath, True, '')
    ufoFontData.readHashMap()

    # Don't need to check the glyph hashes if there aren't any.
    if not ufoFontData.hashMap:
        return allMatch, msgList

    for glyphName, glyphFileName in ufoFontData.getGlyphMap().items():
        hash_entry = ufoFontData.hashMap.get(glyphName, None)
        if not hash_entry:
            continue
        else:
            oldHash = hash_entry[0]

        width, _, outlineXML = ufoFontData.getGlyphXML(
            ufoFontData.glyphDefaultDir, glyphFileName)
        if outlineXML is None:
            continue

        newHash, _ = ufoFontData.buildGlyphHashValue(
            width, outlineXML, glyphName, True)

        if oldHash != newHash:
            allMatch = False
            if len(msgList) < 10:
                msgList.append("Glyph %s seems to have been modified since "
                               "last time checkoutlinesufo processed this "
                               "font." % glyphName)
            elif len(msgList) == 10:
                msgList.append("(additional messages omitted)")

    if doSync:
        fileList = os.listdir(ufoFontData.glyphWriteDir)
        fileList = filter(lambda fileName: fileName.endswith(".glif"),
                          fileList)

        # invert glyphMap
        fileMap = {}
        for glyphName, fileName in ufoFontData.glyphMap.items():
            fileMap[fileName] = glyphName

        for fileName in fileList:
            if fileName in fileMap and (
               fileMap[fileName] in ufoFontData.hashMap):
                continue

            # Either not in glyphMap, or not in hashMap. Exterminate.
            try:
                glyphPath = os.path.join(ufoFontData.glyphWriteDir, fileName)
                os.remove(glyphPath)
                print("Removed outdated file: %s" % glyphPath)
            except OSError:
                print("Cannot delete outdated file: %s" % glyphPath)

    return allMatch, msgList


kAdobeLCALtSuffix = ".adobe.lc.altsuffix"


def cleanUpGLIFFiles(defaultContentsFilePath, glyphDirPath, doWarning=True):
    changed = 0
    contentsFilePath = os.path.join(glyphDirPath, kContentsName)
    # maps glyph names to files.

    with open(contentsFilePath, 'r', encoding='utf-8') as fp:
        contentsDict = plistlib.load(fp)

    # First, delete glyph files that are not in the contents.plist file in
    # the glyphDirPath. In some UFOfont files, we end up with case errors,
    # so we need to check for a lower-case version of the file name.
    fileDict = {}
    for glyphName, fileName in contentsDict.items():
        fileDict[fileName] = glyphName
        lcFileName = fileName.lower()
        if lcFileName != fileName:
            fileDict[lcFileName + kAdobeLCALtSuffix] = glyphName

    fileList = os.listdir(glyphDirPath)
    for fileName in fileList:
        if not fileName.endswith(".glif"):
            continue
        if fileName in fileDict:
            continue
        lcFileName = fileName.lower()
        if (lcFileName + kAdobeLCALtSuffix) in fileDict:
            # glif file exists which has a case-insensitive match to file name
            # entry in the contents.plist file; assume latter is intended, and
            # change the file name to match.
            glyphFilePathOld = os.path.join(glyphDirPath, fileName)
            glyphFilePathNew = os.path.join(glyphDirPath, lcFileName)
            os.rename(glyphFilePathOld, glyphFilePathNew)
            continue

        glyphFilePath = os.path.join(glyphDirPath, fileName)
        os.remove(glyphFilePath)
        if doWarning:
            print("Removing glif file %s that was not in the contents.plist "
                  "file: %s" % (glyphDirPath, contentsFilePath))
        changed = 1

    if defaultContentsFilePath == contentsFilePath:
        return changed

    # Now remove glyphs that are not referenced in the defaultContentsFilePath.
    # Since the processed glyph layer is written with the defcon module,
    # and the default layer may be written by anything, the actual glyph file
    # names may be different for the same UFO glyph. We need to compare by UFO
    # glyph name, not file name.

    with open(defaultContentsFilePath, 'r', encoding='utf-8') as fp:
        defaultContentsDict = plistlib.load(fp)

    fileList = os.listdir(glyphDirPath)
    for fileName in fileList:
        if not fileName.endswith(".glif"):
            continue
        try:
            glyphName = fileDict[fileName]
            if glyphName not in defaultContentsDict:
                glyphFilePath = os.path.join(glyphDirPath, fileName)
                os.remove(glyphFilePath)
                if doWarning:
                    print("Removing glif %s that was not in the "
                          "contents.plist file: %s" % (
                              glyphName, defaultContentsFilePath))
                changed = 1

        except KeyError:
            print("Shouldn't happen %s %s" % (
                glyphName, defaultContentsFilePath))

    return changed


def cleanupContentsList(glyphDirPath, doWarning=True):
    contentsFilePath = os.path.join(glyphDirPath, kContentsName)
    # maps glyph names to files.

    with open(contentsFilePath, 'r', encoding='utf-8') as fp:
        contentsDict = plistlib.load(fp)

    fileDict = {}
    fileList = os.listdir(glyphDirPath)
    for fileName in fileList:
        fileDict[fileName] = 1
    changed = 0

    # now update and write the processed processedGlyphDirPath
    # contents.plist file.
    itemList = list(contentsDict.items())
    for glyphName, fileName in itemList:
        if fileName not in fileDict:
            del contentsDict[glyphName]
            changed = 1
            if doWarning:
                print("Removing contents.plist entry where glif was missing: "
                      "%s, %s, %s" % (glyphName, fileName, glyphDirPath))

    if changed:
        with open(contentsFilePath, 'wb') as fp:
            plistlib.dump(contentsDict, fp)


def validateLayers(ufoFontPath, doWarning=True):
    # Read glyphs/contents.plist file.
    # Delete any glyphs on /glyphs or /processed glyphs which are not in
    # glyphs/contents.plist file. Delete any entries in the contents.plist
    # file which are not in the glyph files. Filter contents list with what's
    # in /processed glyphs: write to process/plist file.' The most common way
    # that this is needed in the AFDKO workflow is if someone kills
    # checkoutlines/checkoutlinesufo or autohint while it is running. Since
    # the program may delete glyphs from the processed layer while running,
    # and the contents.plist file is updated only when the changed font is
    # saved, the contents.plist file in the processed layer ends up referencing
    # glyphs that aren't there anymore. You can also get extra glyphs not in
    # the contents.plist file by several editing workflows.

    # First, clean up the default layer.
    glyphDirPath = os.path.join(ufoFontPath, "glyphs")
    defaultContentsFilePath = os.path.join(
        ufoFontPath, "glyphs", kContentsName)
    # Happens when called on a font which is not a UFO font.
    if not os.path.exists(defaultContentsFilePath):
        return

    # remove glif files not in contents.plist
    cleanUpGLIFFiles(defaultContentsFilePath, glyphDirPath, doWarning)
    # remove entries for glif files that don't exist
    cleanupContentsList(glyphDirPath, doWarning)

    # now for the processed dir.
    glyphDirPath = os.path.join(ufoFontPath, kProcessedGlyphsLayer)
    if not os.path.exists(glyphDirPath):
        return

    # Remove any glif files that are not in both the processed glif directory
    # contents.plist file and the default contents .plist file.
    # This will happen pretty often, as glif files are deleted from the
    # processed glyph layer is their hash differs from the current hash for
    # the glyph in the default layer.
    cleanUpGLIFFiles(defaultContentsFilePath, glyphDirPath, doWarning)
    cleanupContentsList(glyphDirPath, doWarning)


def makeUFOFMNDB(srcFontPath):
    fontInfoPath = os.path.join(srcFontPath, kFontInfoName)  # default
    fiMap, _ = parsePList(fontInfoPath)
    psName = "NoFamilyName-Regular"
    familyName = "NoFamilyName"
    styleName = "Regular"
    try:
        psName = fiMap["postscriptFontName"]
        parts = psName.split("-")
        familyName = parts[0]
        if len(parts) > 1:
            styleName = parts[1]
    except KeyError:
        print("ufotools [Warning] UFO font is missing 'postscriptFontName'")

    try:
        familyName = fiMap["openTypeNamePreferredFamilyName"]
    except KeyError:
        try:
            familyName = fiMap["familyName"]
        except KeyError:
            print("ufotools [Warning] UFO font is missing 'familyName'")

    try:
        styleName = fiMap["openTypeNamePreferredSubfamilyName"]
    except KeyError:
        try:
            styleName = fiMap["styleName"]
        except KeyError:
            print("ufotools [Warning] UFO font is missing 'styleName'")

    fmndbPath = fdkutils.get_temp_file_path()
    parts = []
    parts.append("[%s]" % (psName))
    parts.append("\tf=%s" % (familyName))
    parts.append("\ts=%s" % (styleName))
    parts.append("")
    data = '\n'.join(parts)
    with open(fmndbPath, "w") as fp:
        fp.write(data)
    return fmndbPath


def thresholdAttrGlyph(aGlyph, threshold=0.5):
    """
    Like fontPens.thresholdPen.thresholdGlyph, but preserves some glyph- and
    point-level attributes that are not preserved by that method.
    """
    # preserve glyph-level attributes
    attrnames = ['anchors']
    attrs = {k: getattr(aGlyph, k, None) for k in attrnames if hasattr(aGlyph, k)}  # noqa: E501

    # filter with ThresholdPen into recording pen
    recorder = RecordingPen()
    filterpen = ThresholdPen(recorder, threshold)
    aGlyph.draw(filterpen)
    aGlyph.clear()
    recorder.replay(aGlyph.getPen())

    # restore glyph-level attributes
    for k, v in attrs.items():
        setattr(aGlyph, k, v)

    return aGlyph



"""
Copyright 2019 Adobe
All Rights Reserved.

This software is licensed as OpenSource, under the Apache License, Version 2.0.
This license is available at: http://opensource.org/licenses/Apache-2.0.

--------------------------------------------------------------------------------

This script converts OS/2 Unicode ranges from:
https://docs.microsoft.com/en-us/typography/opentype/spec/os2#ur
(save this section of the page as "os2_ur.txt")

...and Unicode data from:
https://unicode.org/Public/UNIDATA/UnicodeData.txt
(save this file as "UnicodeData.txt")

...into uniblock.h for use in makeotf.
"""
import csv


def _count_chars_in_range(unichars, start, end):
    count = 0
    for unichar in range(start, end + 1):
        if unichar in unichars:
            count += 1
    return count


def _fix_title_casing(name):
    conjunctions = ['And', 'For']
    for conjunction in conjunctions:
        conj_with_spaces = ' ' + conjunction + ' '
        if conj_with_spaces in name:
            name = name.replace(conj_with_spaces, conj_with_spaces.lower())
    return name


def _print_header():
    print(
        '/*\n'
        '   Copyright 2014 Adobe\n'
        '   All Rights Reserved.\n'
        '   This software is licensed as OpenSource, under the Apache License,'
        ' Version 2.0.\n'
        '   This license is available at: '
        'http://opensource.org/licenses/Apache-2.0.\n'
        '*/\n'
        '\n'
        '/*\n'
        '   DO NOT EDIT: this file was generated by the Python script '
        'generate_uniblock.py.\n'
        '   ( c/makeotf/makeotf_lib/utils/generate_uniblock.py )\n'
        '\n'
        '   Unicode block information (struct type: UnicodeBlock). Ordered by '
        'first\n'
        '   Unicode value in range. Ranges do not overlap. '
        'Bit 57 "Non-Plane 0" is \n'
        '   handled separately in map.c (search on SUPP_UV_BITNUM).\n'
        '*/\n'
    )


def _get_unichars():
    unichars = set()
    with open('UnicodeData.txt') as unicodedata_file:
        reader = csv.reader(unicodedata_file, delimiter=';')
        start = None
        for row in reader:
            unichar = int(row[0], 16)
            name = row[1]
            if 'First' in name:
                start = unichar
            elif 'Last' in name:
                for char_in_range in range(start, unichar + 1):
                    unichars.add(char_in_range)
                start = None
            else:
                unichars.add(unichar)
    return unichars


def _get_os2_ur_records():
    # we want to set the corresponding OS/2 Unicode range bits
    # if we see *any* character in a PUA or CJK block
    bits_to_flag_on_any_char = {28, 48, 52, 54, 55, 56, 59, 60, 61, 65, 90}

    unichars = _get_unichars()
    records = list()
    prev_os2_num = 0
    with open('os2_ur.txt') as input_file:
        reader = csv.reader(input_file, delimiter='\t')
        for row in reader:
            if not row[0][0].isdigit():
                row.insert(0, prev_os2_num)
            if row[0] == '123-127':  # reserved stuff at end
                continue
            (os2_num, name, start_end, _) = row
            os2_num = int(os2_num)
            (start, end) = (int(x, 16) for x in start_end.split('-'))
            if os2_num == 57:  # 'Non-Plane 0'
                continue  # this is handled in map.c, search on SUPP_UV_BITNUM
            if os2_num in bits_to_flag_on_any_char:
                count = 1  # we want to flag on *any* characters in this block
            else:
                count = _count_chars_in_range(unichars, start, end)
            name = _fix_title_casing(name)
            record = (start, end, count, os2_num, name)
            records.append(record)
            prev_os2_num = os2_num
    return records


def _print_records(records):
    block_indices = dict()
    blocks_to_remember = {
        'Hiragana': 'HIRAGANA_INX',
        'Katakana': 'KATAKANA_INX',
        'Bopomofo': 'BOPOMOFO_INX',
        'CJK Unified Ideographs': 'CJK_IDEO_INX',
        'Hangul Syllables': 'HANGUL_SYL_INX'}

    for index, record in enumerate(sorted(records)):
        (start, end, count, os2_num, name) = record
        start_str = '0x%04X' % start
        end_str = '0x%04X' % end
        name_str = '"%s"' % name
        print('    { %8s, %8s, %5d, %d, %3d, %d, %-41s },'
              % (start_str, end_str, count, 0, os2_num, 0, name_str))
        if name in blocks_to_remember:
            block_indices[name] = index

    print('')
    print('MOVE THESE VALUES INTO map.c:')
    for name in blocks_to_remember:
        print('#define %-14s %3d' % (
            blocks_to_remember[name], block_indices[name]))


def _main():
    _print_header()
    records = _get_os2_ur_records()
    _print_records(records)


if __name__ == '__main__':
    _main()



# Copyright 2018 Adobe. All rights reserved.

"""
Takes in a TrueType font and looks for a UFO font stored in the same folder.
Uses the UFO's components data to componentize matching TrueType glyphs.
The script only supports components that are not scaled, rotated nor flipped.
"""

import argparse
import os
import sys

from fontTools.ttLib import TTFont, getTableModule
from fontTools.ufoLib.errors import UFOLibError
from defcon import Font

from afdko.fdkutils import get_font_format


__version__ = '0.3.2'


PUBLIC_PSNAMES = "public.postscriptNames"
GOADB_FILENAME = "GlyphOrderAndAliasDB"


class ComponentsData(object):
    def __init__(self):
        self.names = ()
        self.positions = ()
        self.same_advwidth = False

    def add_component(self, name, pos):
        self.names += (name,)
        self.positions += (pos,)


class TTComponentizer(object):
    def __init__(self, ufo, ps_names, input_path, output_path=None):
        self.ufo = ufo
        self.ps_names = ps_names
        self.input_path = input_path
        self.output_path = output_path
        self.composites_data = {}
        self.comp_count = 0

    def componentize(self):
        # Get the composites' info from processing the UFO
        self.get_composites_data()

        self.componentize_ttf()

        plural = "s were" if self.comp_count != 1 else " was"
        print(f"Done! {self.comp_count} glyph{plural} componentized.",
              file=sys.stdout)

    def get_composites_data(self):
        """
        Iterate thru each glyph of a UFO and collect the names and positions
        of all components that make up a composite glyph. The process will
        only collect data for composites that are strictly made with components
        (i.e. no mixed contours-components composites), and whose components
        only have x and y transformations (thus composites with scaled,
        rotated, or flipped componentes will NOT be considered).

        Fills the 'composites_data' dictionary whose keys are composite names
        and whose values are ComponentsData objects.

        NOTE: All glyph names in the returned dictionary are production names.
        """
        for glyph in self.ufo:
            if glyph.components and not len(glyph):
                ttcomps = ComponentsData()
                all_comps_are_basic = True
                for i, comp in enumerate(glyph.components):
                    if comp.transformation[:4] != (1, 0, 0, 1):
                        all_comps_are_basic = False
                        break
                    comp_name = self.ps_names.get(comp.baseGlyph,
                                                  comp.baseGlyph)
                    ttcomps.add_component(comp_name, comp.transformation[4:])
                    if i == 0:
                        ttcomps.same_advwidth = self.check_1st_comp_advwidth(
                            glyph)
                if all_comps_are_basic:
                    glyf_name = self.ps_names.get(glyph.name, glyph.name)
                    self.composites_data[glyf_name] = ttcomps

    def check_1st_comp_advwidth(self, glyph):
        """
        Returns True if the advance width of the composite glyph is the same
        as the advance width of its first component, and False otherwise.
        This information is essential for setting the flag of the composite's
        first component later on.
        """
        return glyph.width == self.ufo[glyph.components[0].baseGlyph].width

    def componentize_ttf(self):
        """
        Loads a TrueType font and iterates thru a dictionary of composites
        data. Remakes some glyphs in the glyf table from being made of
        countours to being made of components.

        Saves the modified font in a new location if an output path was
        provided, otherwise overwrites the original one.

        Updates a count of the glyphs that got componentized.
        """
        font = TTFont(self.input_path)
        glyf_table = font['glyf']

        for gname in self.composites_data:
            if gname not in glyf_table:
                continue
            if not all([cname in glyf_table for cname in self.composites_data[
                    gname].names]):
                continue

            components = self.assemble_components(self.composites_data[gname])

            glyph = glyf_table[gname]
            glyph.__dict__.clear()
            glyph.components = components
            glyph.numberOfContours = -1
            self.comp_count += 1

        if self.output_path:
            font.save(os.path.realpath(self.output_path))
        else:
            font.save(self.input_path)

    @staticmethod
    def assemble_components(comps_data):
        """
        Assemble and return a list of GlyphComponent objects.
        """
        components = []
        for i, cname in enumerate(comps_data.names):
            component = getTableModule('glyf').GlyphComponent()
            component.glyphName = cname
            component.x, component.y = comps_data.positions[i]
            component.flags = 0x4
            if i == 0 and comps_data.same_advwidth:
                component.flags = 0x204
            components.append(component)
        return components


def get_ufo_path(ttf_path):
    """
    Find a UFO font in the same folder as the TT font.
    Returns the UFO's path or None.
    """
    folder_path = os.path.dirname(ttf_path)
    for file_name in sorted(os.listdir(folder_path)):
        if file_name.lower().endswith('.ufo'):
            return os.path.join(folder_path, file_name)
    return None


def get_goadb_path(font_path):
    """
    Find a file named GlyphOrderAndAliasDB. This file can be in the same
    folder as the font, or up-to 3 folders above.
    Returns the GOADB's path or None.
    """
    level = 0
    folder = os.path.dirname(font_path)

    while level <= 3:
        path = os.path.join(folder, GOADB_FILENAME)
        if os.path.exists(path):
            return path
        folder = os.path.dirname(folder)
        level += 1

    return None


def read_txt_file_lines(file_path):
    with open(file_path, "r", encoding="utf-8") as f:
        return f.read().splitlines()


def process_goadb(goadb_path):
    """
    Read a GOADB file and return a dictionary mapping glyph design names to
    glyph production names. The returned mapping may be empty as it will only
    contain entries for glyph names that change from design to production.
    The sctructure of each line of a GOADB file is:
    production_name<tab>design_name<tab>unicode_overrides
    Blank lines and comment lines are alowed in a GOADB file.
    """
    gnames_mapping = {}
    dsgn_names_seen = set()
    prod_names_seen = set()

    for i, line in enumerate(read_txt_file_lines(goadb_path)):
        line = line.strip()
        # skip blank lines
        if not line:
            continue
        # skip comments
        if line.startswith('#'):
            continue

        line_cols = line.split()

        if len(line_cols) < 2:
            print(f"WARNING: Skipped invalid line #{i + 1}: {line}",
                  file=sys.stderr)
            continue

        prod_name, dsgn_name = line_cols[:2]

        if dsgn_name in dsgn_names_seen:
            print(f"WARNING: Skipped duplicate design glyph name '{dsgn_name}'"
                  f" at line #{i + 1}",
                  file=sys.stderr)
            continue

        if prod_name in prod_names_seen:
            print(f"WARNING: Skipped duplicate production glyph name "
                  f"'{prod_name}' at line #{i + 1}",
                  file=sys.stderr)
            continue

        dsgn_names_seen.add(dsgn_name)
        prod_names_seen.add(prod_name)

        # The mapping only needs to contain names that change
        if dsgn_name != prod_name:
            gnames_mapping[dsgn_name] = prod_name

    return gnames_mapping


def get_goadb_names_mapping(ufo_path):
    """
    Assemble a glyph names' mapping dictionary from a GOADB file.
    Returns a dictionary, which can be empty.
    """
    goadb_path = get_goadb_path(ufo_path)

    if not goadb_path:
        return {}

    return process_goadb(goadb_path)


def get_glyph_names_mapping(ufo_path):
    """
    Return a dictionary mapping glyphs' design names to production names.
    This mapping is necessary because the glyph names of the input TTF will
    be production (a.k.a. final) glyph names. In many cases the glyph's design
    name and the production name will be the same, but they may also differ.

    The method will first try to obtain the mapping from the UFO's lib key
    named 'public.postscriptNames'. If this key is not present, the method
    will try to find (and later process) a GlyphOrderAndAliasDB file in the
    folder tree.

    The returned dictionary may be empty, in which case it's assumed that
    there's no difference between the UFO's and the TTF's glyphs' names.

    The UFO object is also returned.
    """
    try:
        ufo = Font(ufo_path)
    except UFOLibError:
        print(f"ERROR: Not a valid UFO font at {ufo_path}",
              file=sys.stderr)
        return None, None

    if PUBLIC_PSNAMES in ufo.lib:
        ufo_names = ufo.lib[PUBLIC_PSNAMES]
        if len(ufo_names) == 0:
            # empty dict; warn
            print(f"WARNING: The contents of {PUBLIC_PSNAMES} is empty. This "
                  "may result in uncomponentized glyphs if final glyph names "
                  "vary from production names.")
        return ufo, ufo_names

    return ufo, get_goadb_names_mapping(ufo_path)


def _validate_font_path(path_str):
    vpath = os.path.abspath(os.path.realpath(path_str))
    if os.path.isfile(vpath) and (get_font_format(vpath) == 'TTF'):
        return vpath
    raise argparse.ArgumentTypeError(
        f"'{path_str}' is not a valid TrueType font file path.")


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-o',
        metavar='OUTPUT_PATH',
        dest='output_path',
        help='path to output the componentized TTF to.'
    )
    parser.add_argument(
        'input_path',
        metavar='FONT',
        type=_validate_font_path,
        help='TTF font file.',
    )
    options = parser.parse_args(args)
    return options


def main(args=None):
    opts = get_options(args)

    # Find UFO file in the same directory
    ufo_path = get_ufo_path(opts.input_path)
    if not ufo_path:
        print(f"ERROR: No UFO font was found for {opts.input_path}",
              file=sys.stderr)
        return 1

    # Get the design->production glyph names mapping, and the UFO object
    ufo, ps_names = get_glyph_names_mapping(ufo_path)
    if not ufo:
        return 1

    ttcomp = TTComponentizer(ufo, ps_names, opts.input_path, opts.output_path)
    ttcomp.componentize()


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 Adobe. All rights reserved.

"""
Generates UFO font instances from a set of master UFO fonts.
It uses the mutatorMath library. The paths to the masters and
instances fonts are specified in the .designspace file.
"""

import argparse
import logging
import os
import shutil
import sys

from fontTools.designspaceLib import (
    DesignSpaceDocument,
    DesignSpaceDocumentError,
)

from defcon import Font
from psautohint.__main__ import main as psautohint
from ufonormalizer import normalizeUFO
from ufoProcessor import build as ufoProcessorBuild

from afdko.checkoutlinesufo import run as checkoutlinesUFO
from afdko.fdkutils import (
    validate_path,
)
from afdko.ufotools import validateLayers


__version__ = '2.4.4'

logger = logging.getLogger(__name__)


DFLT_DESIGNSPACE_FILENAME = "font.designspace"
TEMP_DESIGNSPACE_FILENAME = ".designspace"  # the '.' prefix is intentional
FEATURES_FILENAME = "features.fea"


def filterDesignspaceInstances(dsDoc, options):
    """
    - Filter unwanted instances out of dsDoc as specified by -i option
      (options.indexList), which has already been validated.
    - Promote dsDoc.instance.paths to absolute, referring to original
      dsDoc's location.
    - Remove any existing instance
    - Write the modified doc to a proper temp file
    - Return the path to the temp DS file.
    """

    origDSPath = options.dsPath
    filteredInstances = []

    for idx in sorted(options.indexList):
        instance = dsDoc.instances[idx]
        instance.path = os.path.abspath(
            os.path.realpath(
                os.path.join(
                    origDSPath,
                    instance.path,
                )
            )
        )

        if os.path.exists(instance.path):
            shutil.rmtree(instance.path, ignore_errors=True)

        filteredInstances.append(instance)

    dsDoc.instances = filteredInstances
    # Cannot use a proper tempfile here because on Windows the temporary file
    # may end up being created in a different drive than the original file,
    # which will cause a ValueError when fontTools.designspaceLib tries to
    # compute the relative paths.
    # See https://github.com/fonttools/fonttools/issues/1735
    tmpPath = os.path.join(
        os.path.dirname(origDSPath), TEMP_DESIGNSPACE_FILENAME)
    dsDoc.write(tmpPath)

    return tmpPath


def updateInstance(fontInstancePath, options):
    """
    Run checkoutlinesufo and psautohint, unless explicitly suppressed.
    """
    if options.doOverlapRemoval:
        logger.info("Doing overlap removal with checkoutlinesufo on %s ..." %
                    fontInstancePath)
        co_args = ['-e', '-q', fontInstancePath]
        if options.no_round:
            co_args.insert(0, '-d')
        try:
            checkoutlinesUFO(co_args)
        except (Exception, SystemExit):
            raise

    if options.doAutoHint:
        logger.info("Running psautohint on %s ..." % fontInstancePath)
        ah_args = ['--no-zones-stems', fontInstancePath]
        if options.no_round:
            ah_args.insert(0, '-d')
        try:
            psautohint(ah_args)
        except (Exception, SystemExit):
            raise


def clearCustomLibs(dFont):
    for key in list(dFont.lib.keys()):
        if key not in ['public.glyphOrder', 'public.postscriptNames']:
            del(dFont.lib[key])

    libGlyphs = [g for g in dFont if len(g.lib)]
    for g in libGlyphs:
        g.lib.clear()


def roundSelectedFontInfo(fontInfo):
    """
    'roundGeometry' is false, however, most FontInfo values have to be
    integer, with the exception of:
      - any of the postscript Blue values;
      - postscriptStemSnapH/V;
      - postscriptSlantAngle;

    The Blue values should be rounded to 2 decimal places, with the
    exception of postscriptBlueScale.

    The float values get rounded because most Type1/Type2 rasterizers store
    point and stem coordinates as Fixed numbers with 8 bits; if you pass in
    relative values with more than 2 decimal places, you end up with
    cumulative rounding errors. Other FontInfo float values are stored as
    Fixed number with 16 bits, and can support 6 decimal places.
    """
    for prop_name in fontInfo._properties.keys():
        prop_val = getattr(fontInfo, prop_name)

        if isinstance(prop_val, float):

            if prop_name == 'postscriptBlueScale':
                round_val = round(prop_val, 6)

            elif prop_name in ('postscriptBlueFuzz', 'postscriptBlueShift',
                               'postscriptSlantAngle'):
                round_val = round(prop_val, 2)
            else:
                round_val = int(round(prop_val))

            if (round_val == prop_val) and (prop_val % 1 == 0):
                # defcon does a value comparison before setting a new value
                # on the property, and it will not modify the existing value
                # if the comparison returns True, like e.g. 1 == 1.0
                # Therefore, the existing value is modified by an increment
                # ahead of setting the desired (rounded) value.
                setattr(fontInfo, prop_name, prop_val + 1)
            setattr(fontInfo, prop_name, round_val)

        elif isinstance(prop_val, list) and prop_name.startswith('postscript'):

            for i, val in enumerate(prop_val):
                if isinstance(val, float):
                    prop_val[i] = round(val, 2)


def roundPostscriptBlueScale(fontInfo):
    psbs_str = 'postscriptBlueScale'
    psbs_val = getattr(fontInfo, psbs_str, None)
    if psbs_val:
        setattr(fontInfo, psbs_str, round(psbs_val, 6))


def roundGlyphWidths(dFont):
    for dGlyph in dFont:
        dGlyph.width = int(round(dGlyph.width))


def roundKerningValues(dFont):
    if dFont.kerning:
        keys = dFont.kerning.keys()
        for key in keys:
            dFont.kerning[key] = int(round(dFont.kerning[key]))


def postProcessInstance(fontPath, options):
    dFont = Font(fontPath)
    clearCustomLibs(dFont)

    if options.no_round:
        # '-r/--no-round' option was used but certain values (glyph widths,
        # kerning values, font info values) still need to be rounded because
        # of how they are stored in the final OTF
        roundSelectedFontInfo(dFont.info)
        roundGlyphWidths(dFont)
        roundKerningValues(dFont)
    else:
        # ufoProcessor does not round kerning values nor postscriptBlueScale
        # when 'roundGeometry' = False
        roundPostscriptBlueScale(dFont.info)
        roundKerningValues(dFont)

    dFont.save()


def validateDesignspaceDoc(dsDoc, dsoptions, **kwArgs):
    """
    Validate the dsDoc DesignSpaceDocument object, using supplied dsoptions
    and kwArgs. Raises Exceptions if certain criteria are not met. These
    are above and beyond the basic validations in fontTools.designspaceLib
    and are specific to makeinstancesufo.
    """
    if dsDoc.sources:
        for src in dsDoc.sources:
            if not os.path.exists(src.path):
                raise DesignSpaceDocumentError(
                    f"Source file {src.path} does not exist")
    else:
        raise DesignSpaceDocumentError(
            "Designspace file contains no sources."
        )

    if dsDoc.instances:
        if dsoptions.indexList:
            # bounds check on indexList
            maxinstidx = max(dsoptions.indexList)
            if maxinstidx >= len(dsDoc.instances):
                raise IndexError(f"Instance index {maxinstidx} out-of-range")
    else:
        raise DesignSpaceDocumentError(
            "Designspace file contains no instances."
        )

    if dsoptions.useVarlib:
        axis_map = dict()
        for axis in dsDoc.axes:
            ad = axis.asdict()
            axis_map[ad['name']] = axis

    for i, inst in enumerate(dsDoc.instances):
        if dsoptions.indexList and i not in dsoptions.indexList:
            continue
        for attr_name in ('familyName', 'postScriptFontName', 'styleName'):
            if getattr(inst, attr_name, None) is None:
                logger.warning(
                    f"Instance at index {i} has no "
                    f"'{attr_name.lower()}' attribute.")
        if inst.path is None:
            raise DesignSpaceDocumentError(
                f"Instance at index {i} has no 'filename' attribute.")

        if dsoptions.useVarlib:
            # check for extrapolation, which is not supported by varLib
            # for these only warn, do not raise exception
            for dim, val in inst.location.items():
                axis = axis_map[dim]
                mval = axis.map_backward(val)
                if mval < axis.minimum or mval > axis.maximum:
                    logger.warning("Extrapolation is not supported with varlib"
                                   f" ({inst.familyName} {inst.styleName} "
                                   f"{dim}: {val})")


def collect_features_content(instances, inst_idx_lst):
    """
    Returns a dictionary whose keys are 'features.fea' file paths, and the
    values are the contents of the corresponding file.
    """
    fea_dict = {}
    for i, inst_dscrpt in enumerate(instances):
        if inst_idx_lst and i not in inst_idx_lst:
            continue
        ufo_pth = inst_dscrpt.path
        ufo_pth = os.path.abspath(os.path.realpath(ufo_pth))
        fea_pth = os.path.join(ufo_pth, FEATURES_FILENAME)
        if os.path.isfile(fea_pth):
            with open(fea_pth, 'r') as fp:
                fea_cntnts = fp.read()
            fea_dict[fea_pth] = fea_cntnts
    return fea_dict


def run(options):

    ds_doc = DesignSpaceDocument.fromfile(options.dsPath)

    # can still have a successful read but useless DSD (no sources, no
    # instances, sources non-existent, other conditions
    validateDesignspaceDoc(ds_doc, options)

    copy_features = any(src.copyFeatures for src in ds_doc.sources)
    features_store = {}
    if not copy_features:
        # '<features copy="1"/>' is NOT set in any of masters.
        # Collect the contents of 'features.fea' of any existing
        # instances so that they can be restored later.
        features_store = collect_features_content(ds_doc.instances,
                                                  options.indexList)

    # Set the current dir to the design space dir, so that relative paths in
    # the design space file will work.
    dsDir, dsFile = os.path.split(os.path.abspath(options.dsPath))
    os.chdir(dsDir)
    options.dsPath = dsFile

    if options.indexList:
        dsPath = filterDesignspaceInstances(ds_doc, options)
    else:
        dsPath = dsFile

    newInstancesList = [inst.path for inst in ds_doc.instances]
    newInstancesCount = len(newInstancesList)

    icnt_str = f'instance{"" if newInstancesCount == 1 else "s"}'
    tool_str = "fontTools.varlib" if options.useVarlib else "MutatorMath"
    info_str = f"Building {newInstancesCount} {icnt_str} with {tool_str}..."
    logger.info(info_str)
    ufoProcessorBuild(documentPath=dsPath,
                      outputUFOFormatVersion=options.ufo_version,
                      roundGeometry=(not options.no_round),
                      logger=logger, useVarlib=options.useVarlib)

    # Remove temporary designspace file
    if (dsPath != options.dsPath) and os.path.exists(dsPath):
        os.remove(dsPath)

    logger.info("Built %s instances." % newInstancesCount)
    # Remove glyph.lib and font.lib (except for "public.glyphOrder")
    for instancePath in newInstancesList:
        postProcessInstance(instancePath, options)

    if options.doNormalize:
        logger.info("Applying UFO normalization...")
        for instancePath in newInstancesList:
            normalizeUFO(instancePath, outputPath=None, onlyModified=True,
                         writeModTimes=False)

    if options.doAutoHint or options.doOverlapRemoval:
        logger.info("Applying post-processing...")
        # Apply autohint and checkoutlines, if requested.
        for instancePath in newInstancesList:
            # make new instance font.
            updateInstance(instancePath, options)

    # checkoutlinesufo does ufotools.validateLayers()
    if not options.doOverlapRemoval:
        for instancePath in newInstancesList:
            # make sure that there are no old glyphs left in the
            # processed glyphs folder
            validateLayers(instancePath)

    # The defcon library renames glyphs. Need to fix them again
    if options.doOverlapRemoval or options.doAutoHint:
        for instancePath in newInstancesList:
            if options.doNormalize:
                normalizeUFO(instancePath, outputPath=None, onlyModified=False,
                             writeModTimes=False)

    # Restore the contents of the instances' 'features.fea' files
    for fea_pth, fea_cntnts in features_store.items():
        with open(fea_pth, 'w') as fp:
            fp.write(fea_cntnts)


def _split_comma_sequence(comma_str):
    index_lst = [item.strip() for item in comma_str.split(',')]
    int_set = set()
    for item in index_lst:
        try:
            index = int(item)
        except ValueError:
            raise argparse.ArgumentTypeError(
                f"Invalid integer value: '{item}'")
        if index < 0:
            raise argparse.ArgumentTypeError(
                f"Index values must be greater than zero: '{index}'")
        int_set.add(index)
    return sorted(int_set)


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        '-d',
        '--designspace',
        metavar='PATH',
        dest='dsPath',
        type=validate_path,
        default=DFLT_DESIGNSPACE_FILENAME,
        help='path to design space file\n'
             "Default file name is '%(default)s'"
    )
    parser.add_argument(
        '-a',
        '--no-autohint',
        dest='doAutoHint',
        action='store_false',
        help='do NOT autohint the instances'
    )
    parser.add_argument(
        '-c',
        '--no-remove-overlap',
        dest='doOverlapRemoval',
        action='store_false',
        help="do NOT remove the instances' outline overlaps"
    )
    parser.add_argument(
        '-n',
        '--no-normalize',
        dest='doNormalize',
        action='store_false',
        help='do NOT normalize the instances'
    )
    parser.add_argument(
        '--use-varlib',
        dest='useVarlib',
        action='store_true',
        help='Use varLib instead of MutatorMath'
    )
    parser.add_argument(
        '-r',
        '--no-round',
        action='store_true',
        help='do NOT round coordinates to integer'
    )
    parser.add_argument(
        '-i',
        metavar='INDEX(ES)',
        dest='indexList',
        type=_split_comma_sequence,
        default=[],
        help='index of instance(s) to generate\n'
             'Zero-based comma-separated sequence of integers\n'
             'Example: -i 1,4 (generates 2nd & 5th instances)'
    )
    parser.add_argument(
        '--ufo-version',
        metavar='NUMBER',
        choices=(2, 3),
        default=3,
        type=int,
        help='specify the format version number of the generated UFOs\n'
             'Default: %(default)s'
    )
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    return options


def main(args=None):
    opts = get_options(args)

    try:
        run(opts)
    except Exception as exc:
        logger.error(exc)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import os
import subprocess
import time

import pytest
from differ import main as differ
from runner import main as runner
from test_utils import get_expected_path, get_input_path

TOOL = 'sfntdiff'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

def test_exit_no_option():
    # When ran by itself, 'sfntdiff' prints the usage
    assert subprocess.call([TOOL]) == 1


@pytest.mark.parametrize('arg', ['-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-j', '--bogus'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


@pytest.mark.parametrize('args, txt_filename', [
    ([], 'dflt.txt'),
    (['T'], 'dflt.txt'),  # default diff with timestamp
    (['d', '_0'], 'dflt.txt'),  # level 0 diff
    (['d', '_1'], 'level1.txt'),
    (['d', '_2'], 'level2.txt'),
    (['d', '_3'], 'level3.txt'),
    (['d', '_4'], 'level4.txt'),
])
def test_diff(args, txt_filename):
    if args:
        args.insert(0, '-o')
    actual_path = runner(CMD + ['-s', '-f', 'regular.otf', 'bold.otf'] + args)
    expected_path = get_expected_path(txt_filename)
    assert differ([expected_path, actual_path, '-l', '1-4'])


def test_diff_otf_vs_ttf_bug626():
    actual_path = runner(CMD + ['-s', '-f', 'SourceSerifPro-It.otf',
                                            'SourceSerifPro-It.ttf'])
    expected_path = get_expected_path('bug626.txt')
    assert differ([expected_path, actual_path, '-l', '1-4'])


def test_not_font_files():
    stderr_path = runner(CMD + ['-s', '-e', '-f', 'not_a_font_1.otf',
                                                  'not_a_font_2.otf'])
    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert b'[WARNING]: unsupported/bad file' in output
    assert b'not_a_font_1.otf] (ignored)' in output
    assert b'not_a_font_2.otf] (ignored)' in output


def test_date_and_time1():
    """This test checks the date and time at the top of the report."""
    args = ['-o', 'd', '_0']
    actual_path = runner(CMD + ['-s', '-f', 'regular.otf', 'bold.otf'] + args)
    with open(actual_path) as output_file:
        line = output_file.readline().strip()
        now = time.time()
        file_time = time.mktime(time.strptime(line, '%a %b %d %H:%M:%S %Y'))
        hours_diff = abs(now - file_time) / 3600
        assert(hours_diff < 1)


def test_date_and_time2():
    """This test checks the two different ways of getting the date and time
       when comparing the dates of two fonts using the -T option.
       If the font has a 'head' table, the code uses the 'head.created' time,
       but if does not, e.g. in the case of a CEF font, then the code uses
       stat() to get the file's modification time."""
    args = ['-o', 'T']
    actual_path = runner(CMD + ['-s', '-f', 'regular.otf', 'font.cef'] + args)

    cef_path = get_input_path('font.cef')
    cef_secs_since_epoch = os.path.getmtime(cef_path)
    cef_localtime = time.localtime(cef_secs_since_epoch)
    cef_time_str = time.strftime('%a %b %d %H:%M:%S %Y', cef_localtime)

    with open(actual_path) as output_file:
        text = output_file.read()
        assert('regular.otf\tMon Aug 31 14:44:59 2015' in text)
        assert('font.cef\t' + cef_time_str in text)



import argparse
import pytest

from differ import (
    main as differ,
    get_options,
    _compile_regex,
    _convert_to_int,
    _split_linenumber_sequence,
    _validate_path,
    _get_path_kind,
)
from test_utils import get_input_path


@pytest.mark.parametrize('v_arg', ['', 'v', 'vv', 'vvv'])
def test_log_level(v_arg):
    fp1, fp2 = get_input_path('file1'), get_input_path('file2')
    v_val = len(v_arg)
    arg = [] if not v_val else [f'-{v_arg}']
    opts = get_options([fp1, fp2] + arg)
    assert opts.verbose == v_val


@pytest.mark.parametrize('fname, result', [
    ('folder_copy', True),
    ('folder_mod_file', False),
    ('folder_less_file', False),
    ('folder_more_file', False),
])
@pytest.mark.parametrize('difmod', [[], ['-m', 'bin']])
def test_compare_dir_contents(fname, result, difmod):
    fp1, fp2 = get_input_path('folder'), get_input_path(fname)
    assert differ([fp1, fp2] + difmod) is result


@pytest.mark.parametrize('difmod', [[], ['-m', 'bin']])
def test_compare_files(difmod):
    fp1, fp2 = get_input_path('file1'), get_input_path('file2')
    assert differ([fp1, fp2] + difmod) is False


def test_file_encoding():
    fp1, fp2 = get_input_path('file1'), get_input_path('file2_macroman')
    with pytest.raises(SystemExit) as exc_info:
        differ([fp1, fp2])
    assert exc_info.value.code == 1
    assert differ([fp1, fp2, '-e', 'macroman']) is False


def test_paths_not_same_kind():
    fp1, fp2 = get_input_path('file1'), get_input_path('folder')
    with pytest.raises(SystemExit) as exc_info:
        differ([fp1, fp2])
    assert exc_info.value.code == 2


def test_validate_path():
    with pytest.raises(argparse.ArgumentTypeError):
        _validate_path(get_input_path('not_a_file_or_folder'))


@pytest.mark.parametrize('pth, result', [
    ('not_a_file_or_folder', 'invalid'),
    ([], 'invalid'),
])
def test_get_path_kind(pth, result):
    assert _get_path_kind(pth) == result


@pytest.mark.parametrize('regex', ['', '^(ab'])
def test_compile_regex(regex):
    with pytest.raises(argparse.ArgumentTypeError):
        _compile_regex(regex)


@pytest.mark.parametrize('string', ['', 'a'])
def test_convert_to_int(string):
    with pytest.raises(argparse.ArgumentTypeError):
        _convert_to_int(string)


@pytest.mark.parametrize('str_seq, tup_seq', [
    ('10', (10,)),
    ('11,13,19', (11, 13, 19)),
    ('13-16', (13, 14, 15, 16)),
    ('15+4', (15, 16, 17, 18, 19)),
    ('10-10,5+4,2,3-5', (2, 3, 4, 5, 6, 7, 8, 9, 10)),
])
def test_split_linenumber_sequence(str_seq, tup_seq):
    assert _split_linenumber_sequence(str_seq) == tup_seq


@pytest.mark.parametrize('str_seq', ['9*7', '7_9', '9-7'])
def test_split_linenumber_sequence_error(str_seq):
    with pytest.raises(argparse.ArgumentTypeError):
        _split_linenumber_sequence(str_seq)



# Copyright 2014 Adobe. All rights reserved.

"""
Extract all OpenType fonts from the parent OpenType Collection font.
"""

import argparse
import logging
import os
import sys

from fontTools.ttLib import sfnt, TTFont

from afdko.fdkutils import (
    get_font_format,
    validate_path,
)

__version__ = '2.0.0'

logger = logging.getLogger(__name__)


def get_psname(font):
    """
    Returns the font's nameID 6 (a.k.a. PostScript name).
    Returns None if the font has no nameID 6, if nameID 6 is an empty
    string, or if the font has no 'name' table.
    """
    if 'name' in font:
        psname = font['name'].getDebugName(6)
        if psname:
            return psname


def run(options):
    ttc_path = options.ttc_path

    with open(ttc_path, 'rb') as fp:
        num_fonts = sfnt.SFNTReader(fp, fontNumber=0).numFonts

    if options.report:
        unique_tables = set()
        print(f'Input font: {os.path.basename(ttc_path)}\n')

    for ft_idx in range(num_fonts):
        font = TTFont(ttc_path, fontNumber=ft_idx, lazy=True)

        psname = get_psname(font)
        if psname is None:
            ttc_name = os.path.splitext(os.path.basename(ttc_path))[0]
            psname = f'{ttc_name}-font{ft_idx}'

        ext = '.otf' if font.sfntVersion == 'OTTO' else '.ttf'
        font_filename = f'{psname}{ext}'

        if options.report:
            # this code is based on fontTools.ttx.ttList
            reader = font.reader
            tags = sorted(reader.keys())
            print(f'Font {ft_idx}: {font_filename}')
            print('    tag ', '   checksum', '   length', '   offset')
            print('    ----', ' ----------', ' --------', ' --------')
            for tag in tags:
                entry = reader.tables[tag]
                checksum = int(entry.checkSum)
                if checksum < 0:
                    checksum += 0x100000000
                checksum = f'0x{checksum:08X}'
                tb_info = (tag, checksum, entry.length, entry.offset)
                if tb_info not in unique_tables:
                    print('    {:4s}  {:10s}  {:8d}  {:8d}'.format(*tb_info))
                    unique_tables.add(tb_info)
                else:
                    print(f'    {tag:4s}  - shared -')
            print()
        else:
            save_path = os.path.join(os.path.dirname(ttc_path), font_filename)
            font.save(save_path)
            logger.info(f'Saved {save_path}')

        font.close()


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        '-r',
        '--report',
        action='store_true',
        help="report the TTC's fonts and tables (no files are written)",
    )
    parser.add_argument(
        'ttc_path',
        metavar='TTC FONT',
        type=validate_path,
        help='path to TTC font',
    )
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    if get_font_format(options.ttc_path) != 'TTC':
        parser.error('The input file is not an OpenType Collection font.')

    return options


def main(args=None):
    opts = get_options(args)

    try:
        run(opts)
    except Exception as exc:
        logger.error(exc)
        return 1


if __name__ == "__main__":
    sys.exit(main())
